# 786
# Aldy source: common.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Iterable

import pkg_resources
import re
import time
import pprint
import logbook
import textwrap
import collections


PROTEINS = {
    "TTT": "F",
    "CTT": "L",
    "ATT": "I",
    "GTT": "V",
    "TTC": "F",
    "CTC": "L",
    "ATC": "I",
    "GTC": "V",
    "TTA": "L",
    "CTA": "L",
    "ATA": "I",
    "GTA": "V",
    "TTG": "L",
    "CTG": "L",
    "ATG": "M",
    "GTG": "V",
    "TCT": "S",
    "CCT": "P",
    "ACT": "T",
    "GCT": "A",
    "TCC": "S",
    "CCC": "P",
    "ACC": "T",
    "GCC": "A",
    "TCA": "S",
    "CCA": "P",
    "ACA": "T",
    "GCA": "A",
    "TCG": "S",
    "CCG": "P",
    "ACG": "T",
    "GCG": "A",
    "TAT": "Y",
    "CAT": "H",
    "AAT": "N",
    "GAT": "D",
    "TAC": "Y",
    "CAC": "H",
    "AAC": "N",
    "GAC": "D",
    "TAA": "X",
    "CAA": "Q",
    "AAA": "K",
    "GAA": "E",
    "TAG": "X",
    "CAG": "Q",
    "AAG": "K",
    "GAG": "E",
    "TGT": "C",
    "CGT": "R",
    "AGT": "S",
    "GGT": "G",
    "TGC": "C",
    "CGC": "R",
    "AGC": "S",
    "GGC": "G",
    "TGA": "X",
    "CGA": "R",
    "AGA": "R",
    "GGA": "G",
    "TGG": "W",
    "CGG": "R",
    "AGG": "R",
    "GGG": "G",
}
"""dict[str, str]: Codon table (stop codon is X)."""


REV_COMPLEMENT = {"A": "T", "T": "A", "C": "G", "G": "C"}
"""dict[str, str]: Reverse-complement DNA table."""


log = logbook.Logger("Aldy")
"""Default console logger."""


SOLUTION_PRECISION = 1e-2
"""float: Solution precision (all values whose difference is less than
   SOLUTION_PRECISION are considered equal)"""


class AldyException(Exception):
    """
    Aldy exception class.
    """

    pass


class GRange(collections.namedtuple("GRange", ["chr", "start", "end"])):
    """
    A range within the reference genome (e.g. chr22:10-20).
    Immutable.

    Attributes:
        chr (str): Chromosome identifier.
        start (int): Start position of the interval.
        end (int): End position of the interval.

    Notes:
        Has custom printer (``__str__``).
    """

    def samtools(self, pad_left=500, pad_right=1, prefix="") -> str:
        """
        Samtools-compatible region representation (e.g. chr1:10-20).

        Returns:
            str
        """
        return "{}:{}-{}".format(
            prefix + self.chr, self.start - pad_left, self.end + pad_right
        )

    def __str__(self):
        return self.samtools(0, 0, "")


# Aldy auxiliaries


def allele_name(x: str) -> str:
    """
    Returns:
        str: Major allele number of the star-allele name (e.g. ``'12A'`` -> ``12``).
    """
    if "*" in x:
        x = x.split("*", maxsplit=1)[1]
    return x.replace("/", "_")


def freezekey(x):  # hashing for dictionaries
    return tuple(i[1] for i in sorted(x[0].items())) + tuple(
        i[1] for i in sorted(x[1].items())
    )


def rev_comp(seq: str) -> str:
    """
    Returns:
        str: Reverse-complemented DNA sequence.
    """

    return "".join([REV_COMPLEMENT.get(x, x) for x in seq[::-1]])


def seq_to_amino(seq: str) -> str:
    """
    Returns:
        str: Protein sequence formed from the provided DNA sequence.
    """

    return "".join(
        PROTEINS[seq[i : i + 3]] for i in range(0, len(seq) - len(seq) % 3, 3)
    )


# Language auxiliaries


def sorted_tuple(x: Iterable) -> tuple:
    """
    Sort a tuple.
    """
    return tuple(sorted(x))


def td(s: str) -> str:
    """
    Abbreviation for textwrap.dedent. Useful for stripping indentation
    in multi-line docstrings.
    """
    return textwrap.dedent(s)


def timing(f):  # pragma: no cover
    """
    Decorator for timing a function.
    Prints the time spent in the function after once it is completed.

    Usage::

        @timing

    (without any parameters).
    """

    def wrap(*args, **kwargs):
        time1 = time.time()
        ret = f(*args, **kwargs)
        time2 = time.time()
        log.warn("Time needed: ({:.1f})", time2 - time1)
        return ret

    return wrap


def pp(x) -> str:
    """
    Returns:
        str: Pretty-printed variable string.
    """
    return pprint.pformat(x)


def pr(x):
    """
    Pretty-print a variable to stdout.
    """
    pprint.pprint(x)


def script_path(key: str) -> str:
    """
    Args:
        key (str): resource to be extracted.
        Specify as ``path/file`` (e.g. ``aldy.resources/test.txt``).

    Returns:
        str: Full path of the resource.

    Raises:
        :obj:`aldy.common.AldyException`.
    """
    components = key.split("/")
    if len(components) < 2:
        raise AldyException(f'"{key}"" is not valid resource name')
    return pkg_resources.resource_filename(components[0], "/".join(components[1:]))


def colorize(text: str, color: str = "green") -> str:
    """
    Returns:
        str: Colorized string (on xterm-compatible terminals) with a given color.
    """
    return logbook._termcolors.colorize(color, text)


def parse_cn_region(cn_region):
    if cn_region is not None:
        r = re.match(r"^(.+?):(\d+)-(\d+)$", cn_region)
        if not r:
            raise AldyException(
                f"Parameter --cn-neutral={cn_region} cannot be parsed. "
                + "Must be chr:start-end (where start and end are numbers)"
            )
        ch = r.group(1)
        if ch.startswith("chr"):
            ch = ch[3:]
        return GRange(ch, int(r.group(2)), int(r.group(3)))
    return None


class JsonDict(dict):
    def __getitem__(self, key):
        if key not in self:
            self[key] = JsonDict()
        return self.get(key)


json = JsonDict()



# 786
# Aldy source: solutions.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import collections
from dataclasses import dataclass, field
from natsort import natsorted
from typing import List, Dict, Optional
from .gene import Gene, Mutation


@dataclass
class CNSolution:
    r"""
    Valid copy-number configuration assignment.

    :param gene: Gene instance.
    :param score: ILP model objective score (0 for user-provided solutions).
    :param solution:
        Copy-number configurations mapped to the corresponding copy number
        (e.g. ``{1: 2}`` means that there are two copies of \*1 configuration).
    :param region_cn: Gene region copy numbers inferred by this solution.

    .. note:: Has custom printer (``__str__``).
    """

    gene: Gene
    score: float
    solution: Dict[str, int]
    region_cn: List[Dict[str, int]]

    def __init__(self, gene: Gene, score: float, solution: List[str]):
        self.region_cn = [
            {r: 0 for r in gene.regions[0]} for _ in gene.cn_configs["1"].cn
        ]
        for conf in solution:
            for gi, g in enumerate(gene.cn_configs[conf].cn):
                for r in g:
                    self.region_cn[gi][r] += g[r]
        self.gene = gene
        self.score = score
        self.solution = collections.Counter(solution)

    def __hash__(self):
        return str(self).__hash__()

    def position_cn(self, pos: int) -> float:
        """:return: Copy number at the locus ``pos``."""
        r = self.gene.region_at(pos)
        return self.region_cn[r[0]][r[1]] if r else 0

    def _solution_nice(self):
        return ",".join(f"{v}x*{k}" for k, v in natsorted(self.solution.items()))

    def __str__(self):
        return "CNSol[{:.2f}; sol=({}); cn={}]".format(
            self.score,
            self._solution_nice(),
            "|".join(
                "".join(str(self.region_cn[g][r]) for r in self.gene.regions[0])
                for g, _ in enumerate(self.region_cn)
            ),
        )

    def max_cn(self):
        """Maximum copy-number in this solution"""
        return sum(self.solution.values())


@dataclass
class SolvedAllele:
    """
    Solved star-allele assignment.

    :param gene: Gene instance.
    :param major: Major star-allele identifier.
    :param minor: Minor star-allele identifier. Can be None.
    :param added:
        Mutations that are added to the star-allele
        (present in the allele database definition).
    :param missing:
        Mutations that are omitted from the star-allele
        (present in the allele database definition but not here).

    .. note:: Has custom printer (``__str__``).
    """

    gene: Gene
    major: str
    minor: Optional[str] = None
    added: List[Mutation] = field(default_factory=list)
    missing: List[Mutation] = field(default_factory=list)

    def mutations(self):
        m = self.gene.alleles[self.major].func_muts
        m |= self.gene.alleles[self.major].minors[self.minor].neutral_muts
        m |= set(self.added)
        m -= set(self.missing)
        return m

    def major_repr(self):
        """
        Pretty-formats major star-allele name.
        """
        return "*{}{}".format(
            self.major,
            "".join(
                " +" + self.gene.get_rsid(m)
                for m in sorted(
                    m for m in self.added if self.gene.is_functional(m, False)
                )
            ),
        )

    def __str__(self):
        """
        Pretty-formats minor star-allele name.
        """
        return "*{}{}{}".format(
            self.minor if self.minor else self.major,
            "".join(" +" + self.gene.get_rsid(m) for m in sorted(self.added)),
            "".join(" -" + self.gene.get_rsid(m) for m in sorted(self.missing)),
        )

    def __hash__(self):
        return hash(
            (self.gene.name, self.major, self.minor, *self.added, *self.missing)
        )


@dataclass
class MajorSolution:
    r"""
    Valid major star-allele assignment.

    :param score: ILP model objective score.
    :param solution:
        Major star-alleles and the corresponding copy numbers
        (e.g. ``{1: 2}`` means that we have two copies of \*1).
    :param cn_solution: Copy-number solution that was used to assign major star-alleles.
    :param added:
        List of added mutations. Will be assigned to :obj:`SolvedAllele` in the
        minor star-allele calling step if phasing is enabled.

    .. note:: Has custom printer (``__str__``).
    """

    score: float
    solution: Dict[SolvedAllele, int]
    cn_solution: CNSolution
    added: List[Mutation]

    def _solution_nice(self):
        x = ", ".join(
            f"{v}x{s}"
            for s, v in natsorted(self.solution.items(), key=lambda x: x[0].major)
        )
        y = ", ".join(self.cn_solution.gene.get_rsid(m) for m in sorted(self.added))
        return " & ".join([x, y] if y else [x])

    def __str__(self):
        return (
            f"MajorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"cn={self.cn_solution}"
        )

    def __hash__(self):
        return str(self).__hash__()


@dataclass
class MinorSolution:
    """
    Valid minor star-allele assignment.

    :param score: ILP model objective score.
    :param solution:
        List of solved minor star-alleles.
        Modifications to minor alleles are represented in :obj:`SolvedAllele` format.
    :param major_solution:
        Major star-allele solution used for calculating minor star-allele assignments.

    .. note:: Has custom printer (``__str__``).
    """

    score: float
    solution: List[SolvedAllele]
    major_solution: MajorSolution

    def _solution_nice(self):
        return ", ".join(
            str(s) for s in natsorted(self.solution, key=lambda x: x.minor)
        )

    def __str__(self):
        return (
            f"MinorSol[{self.score:.2f}; "
            + f"sol=({self._solution_nice()}); "
            + f"major={self.major_solution}"
        )

    def get_diplotype(self):
        try:
            return self.diplotype
        except AttributeError:
            self.diplotype = ([], [])
            return self.diplotype

    def set_diplotype(self, d):
        self.diplotype = d

    def get_major_name(self, i):
        gene = self.major_solution.cn_solution.gene
        if i == -1:
            return gene.deletion_allele()
        n = str(self.solution[i].major).split("#")[0:1]
        for m in sorted(self.solution[i].added):
            if gene.is_functional(m, infer=False):
                n.append(gene.get_rsid(m))
        return "+".join(n)

    def get_minor_name(self, i, legacy=False):
        gene = self.major_solution.cn_solution.gene
        if i == -1:
            return gene.deletion_allele()

        m = gene.alleles[self.solution[i].major]
        t = [mi for mi in m.minors if mi == self.solution[i].minor]
        assert len(t) == 1
        n = [m.minors[t[0]].alt_name if legacy and m.minors[t[0]].alt_name else t[0]]
        for m in sorted(self.solution[i].added):
            n.append("+" + gene.get_rsid(m))
        for m in sorted(self.solution[i].missing):
            n.append("-" + gene.get_rsid(m))
        return " ".join(n)

    def get_major_diplotype(self):
        return " / ".join(
            " + ".join(f"*{self.get_major_name(i)}" for i in d) for d in self.diplotype
        )

    def get_minor_diplotype(self, legacy=False):
        return " / ".join(
            " + ".join(f"[*{self.get_minor_name(i, legacy)}]" for i in d)
            for d in self.diplotype
        )



# 786
# Aldy source: conftest.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import pytest
import ast

from aldy.gene import Gene
from aldy.common import script_path


@pytest.fixture
def toy_gene():
    return Gene(script_path("aldy.tests.resources/toy.yml"))


@pytest.fixture
def real_gene():
    return Gene(script_path("aldy.resources.genes/cyp2d6.yml"))


def pytest_addoption(parser):
    parser.addoption("--samples", action="store", default=None)
    parser.addoption("--solvers", action="store", default=None)


def pytest_generate_tests(metafunc):
    path = metafunc.config.getoption("samples")
    solvers = metafunc.config.getoption("solvers")
    if solvers is None:
        solvers = "any"
    solvers = solvers.split(",")

    if "solver" in metafunc.fixturenames:
        metafunc.parametrize("solver", solvers)
    if "real_sample" in metafunc.fixturenames and not path:
        metafunc.parametrize("real_sample", [])
    elif "real_sample" in metafunc.fixturenames and path:

        def read_data(path):
            with open(script_path(path)) as f:
                data = f.read()
                data = ast.literal_eval(data)
                return data

        samples = []
        for file, data in read_data("aldy.tests.resources/data-pgx1.json").items():
            samples.append(
                (
                    data,
                    f"{path}/cdc/pgrnseq-v1/bams/{file}.cram",
                    "pgrnseq-v1",
                    f"{path}/cram-genome.fa",
                )
            )
        for file, data in read_data("aldy.tests.resources/data-pgx2.json").items():
            samples.append(
                (
                    data,
                    f"{path}/baylor/pgrnseq-v2/bams/{file}.cram",
                    "pgrnseq-v2",
                    f"{path}/cram-genome.fa",
                )
            )
        for file, data in read_data("aldy.tests.resources/data-illumina.json").items():
            samples.append(
                (data, f"{path}/1000genomes-illumina/bams/{file}.bam", "illumina", None)
            )
        metafunc.parametrize("real_sample", samples)



# 786
# Aldy source: sam.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional, Any, Set

import pysam
import os
import os.path
import re
import yaml
import gzip
import struct

from dataclasses import dataclass
from collections import defaultdict
from natsort import natsorted
from .common import log, GRange, AldyException, script_path
from .coverage import Coverage
from .gene import Gene, Mutation


CIGAR2CODE: Dict[int, int] = {ord(y): x for x, y in enumerate("MIDNSHP=XB")}
"""CIGAR characters for an ASCII ordinal of CIGAR character"""

CIGAR_REGEX = re.compile(r"(\d+)([MIDNSHP=XB])")
"""Regex that matches valid CIGAR strings"""


DEFAULT_CN_NEUTRAL_REGION = {
    "hg19": GRange("22", 42547463, 42548249),
    "hg38": GRange("22", 42151472, 42152258),
}
"""Default copy-number neutral region (exon 4-6 of the CYP2D8 gene)"""


@dataclass
class Sample:
    """
    Interface for reading SAM/BAM/CRAM files that parses and stores read alignments.
    """

    coverage: Coverage
    sample_name: str
    _prefix: str
    _insertion_sites: Set[Mutation]
    _insertion_counts: Dict[Tuple[int, int], int]
    _insertion_reads: Dict[Mutation, Dict[Any, int]]
    _multi_sites: Dict[int, str]
    _dump: Tuple[Dict[int, int], List[Tuple[Tuple[int, int, int], List[Mutation]]]]

    def __init__(
        self,
        gene: Gene,
        sam_path: Optional[str] = None,
        threshold: float = 0.5,
        profile: Optional[str] = None,
        reference: Optional[str] = None,
        cn_region: Optional[GRange] = None,
        debug: Optional[str] = None,
        vcf_path: Optional[str] = None,
    ):
        """
        Initialize a :obj:`Sample` object.

        :param sam_path: Path to a SAM/BAM/CRAM file.
        :param gene: Gene instance.
        :param threshold:
            Threshold for filtering out low quality mutations. Ranges from 0 to 1.
            Check :obj:`aldy.coverage.Coverage` for more information.
        :param profile:
            Profile specification (e.g. 'prgnseq-v1') or a profile SAM/BAM file.
        :param reference:
            Reference genome for reading CRAM files.
            Default is None.
        :param cn_region:
            Copy-number neutral region to be used for coverage rescaling.
            If None, profile loading and coverage rescaling will be skipped
            (and Aldy will require a ``--cn`` parameter to be user-provided).
            Default is ``DEFAULT_CN_NEUTRAL_REGION``.
        :param debug:
            If set, create a "`debug`.dump" file for debug purposes.
            Default is None.

        :raise: :obj:`aldy.common.AldyException` if the average coverage of the
                copy-number neutral region is too low (less than 2).
        """

        # Get the list of indel sites that should be corrected
        # TODO: currently uses only functional indels;
        #       other indels should be corrected as well
        self._dump = defaultdict(int), []
        self._insertion_sites = {
            m for a in gene.alleles.values() for m in a.func_muts if m.op[:3] == "ins"
        }
        self._insertion_counts: Dict[Tuple[int, int], int] = defaultdict(int)
        self._insertion_reads: Dict[Mutation, Dict[Any, int]] = {
            m: defaultdict(int) for m in self._insertion_sites
        }
        self._multi_sites = {
            m.pos: m.op
            for _, a in gene.alleles.items()
            for m in a.func_muts
            if ">" in m.op and len(m.op) > 3
        }

        assert vcf_path or (sam_path and profile)
        if vcf_path:
            try:
                norm, muts = self._load_vcf(vcf_path, gene)
            except ValueError:
                raise AldyException(f"VCF {vcf_path} is not indexed")
        elif sam_path and sam_path[-5:] == ".dump":
            norm, muts = self._load_dump(sam_path)
        else:
            assert sam_path
            norm, muts = self._load_sam(sam_path, gene, reference, cn_region, debug)
        self._make_coverage(gene, norm, muts, threshold)

        if cn_region:
            assert profile
            self.detect_cn(gene, profile, cn_region)
        log.debug("[sam] avg_coverage= {:.1f}x", self.coverage.average_coverage())
        if cn_region and self.coverage.diploid_avg_coverage() < 2:
            raise AldyException(
                "The average coverage of the sample is too low ({:.1f}).".format(
                    self.coverage.diploid_avg_coverage()
                )
            )

    def _load_sam(
        self,
        sam_path: str,
        gene: Gene,
        reference: Optional[str] = None,
        cn_region: Optional[GRange] = None,
        debug: Optional[str] = None,
    ):
        """
        Load the read, mutation and coverage data from a SAM/BAM/CRAM file.

        :param sam_path: Path to a SAM/BAM/CRAM file.
        :param gene: Gene instance.
        :param threshold:
            Threshold for filtering out low quality mutations. Ranges from 0 to 1.
            Check :obj:`aldy.coverage.Coverage` for more information.
        :param reference:
            Reference genome for reading CRAM files.
            Default is None.
        :param cn_region:
            Copy-number neutral region to be used for coverage rescaling.
            If None, profile loading and coverage rescaling will be skipped
            (and Aldy will require a ``--cn`` parameter to be user-provided).
            Default is ``DEFAULT_CN_NEUTRAL_REGION`` (CYP2D8).
        :param debug:
            If set, create a "`debug`.dump" file for debug purposes.
            Default is None.

        :raise: :obj:`aldy.common.AldyException` if a BAM/CRAM file lacks an index.
        """

        log.debug("[sam] path= {}", os.path.abspath(sam_path))
        self.sample_name = os.path.basename(sam_path).split(".")[0]

        norm: dict = defaultdict(int)
        muts: dict = defaultdict(int)

        # Try to read CN-neutral region if there is an index (fast)
        # If not, the main loop will catch it later on
        def read_cn_read(read):
            """
            Check is pysam.AlignmentSegment valid CN-neutral read,
            and if so, parse it
            """
            start = read.reference_start
            if read.cigartuples is None or (
                (read.flag & 1) and (read.flag & 0x40) == 0
            ):
                return
            for op, size in read.cigartuples:
                if op in [0, 7, 8, 2]:
                    for i in range(size):
                        self._dump[0][start + i] += 1
                    start += size

        with pysam.AlignmentFile(sam_path, reference_filename=reference) as sam:
            # Check do we have proper index to speed up the queries
            try:
                sam.check_index()
            except AttributeError:
                pass  # SAM files do not have an index. BAMs might also lack it
            except ValueError:
                raise AldyException(f"File {sam_path} has no index")

            self._prefix = _chr_prefix(gene.chr, [x["SN"] for x in sam.header["SQ"]])

            # Set it to _fetched_ if a CN-neutral region is user-provided.
            # Then read the CN-neutral region.
            is_cn_region_fetched = cn_region is None
            if cn_region and sam.has_index():
                log.debug("[sam] has_index= True")
                for read in sam.fetch(region=cn_region.samtools(prefix=self._prefix)):
                    read_cn_read(read)
                is_cn_region_fetched = True

            # Fetch the reads
            for read in sam.fetch(
                region=gene.get_wide_region().samtools(prefix=self._prefix)
            ):
                # If we haven't obtained CN-neutral region so far, do it now
                if (
                    not is_cn_region_fetched
                    and cn_region
                    and _in_region(cn_region, read, self._prefix)
                ):  # type: ignore
                    read_cn_read(read)

                r = self._parse_read(read, gene, norm, muts)
                if r and debug:
                    self._dump[1].append(r)
        if debug:
            self._dump_alignments(debug)
        return norm, muts

    def _load_vcf(self, vcf_path: str, gene: Gene):
        """
        Load the read, mutation and coverage data from a VCF file.
        """

        log.debug("[vcf] path= {}", os.path.abspath(vcf_path))

        norm = {
            p: 20
            for p in range(
                gene.get_wide_region().start - 500, gene.get_wide_region().end + 1
            )
        }
        muts: dict = defaultdict(int)

        def get_mut(pos, ref, alt):
            off = 0
            while off < len(ref) and off < len(alt) and ref[off] == alt[off]:
                off += 1
            if len(ref) - off == 1 and len(alt) - off == 1:
                if alt[off] == gene[off + pos]:
                    return off + pos, "_"
                return off + pos, f"{gene[off + pos]}>{alt[off]}"
            elif len(ref) > len(alt) and len(alt) - off == 0:
                return off + pos, f"del{gene[off + pos : pos + len(ref)]}"
            elif len(ref) < len(alt) and len(ref) - off == 0:
                return off + pos, f"ins{alt[off:]}"
            else:
                log.trace(f"[sam] ignoring {pos}: {ref}->{alt}")
                return pos, None

        with pysam.VariantFile(vcf_path) as vcf:
            self._prefix = _chr_prefix(gene.chr, list(vcf.header.contigs))

            samples = list(vcf.header.samples)
            self.sample_name = sample = samples[0]
            if len(samples) > 1:
                log.warn("WARNING: Multiple VCF samples found; using the first one.")
            log.info("Using VCF sample {}", sample)
            for read in vcf.fetch(
                region=gene.get_wide_region().samtools(prefix=self._prefix)
            ):
                g = sorted(y for y in read.samples[sample]["GT"] if y is not None)
                if len(g) != 2 or gene[read.pos - 1] == "N":
                    continue  # ignore polyploid and incomplete cases
                dump_arr = {}
                if len(read.ref) == 1 and read.ref != gene[read.pos - 1]:
                    hgvs = [(read.pos - 1, f"{gene[read.pos - 1]}>{read.ref}")]
                else:
                    hgvs = [(read.pos - 1, "_")]
                hgvs += [get_mut(read.pos - 1, read.ref, a) for a in read.alleles[1:]]
                for gt in g:
                    pos, op = hgvs[gt]
                    if op == "_":
                        continue
                    muts[pos, op] += 10
                    norm[pos] -= 10
                    dump_arr[pos] = op

                for pos, op in self._multi_sites.items():
                    if pos not in dump_arr:
                        continue
                    l, r = op.split(">")
                    if all(
                        dump_arr.get(pos + p, "-") == f"{l[p]}>{r[p]}"
                        for p in range(len(l))
                        if l[p] != "."
                    ):
                        for p in range(len(l)):
                            if l[p] != ".":
                                muts[pos + p, f"{l[p]}>{r[p]}"] -= 10
                                if p:
                                    norm[pos + p] += 10
                        muts[pos, op] += 10
        return norm, muts

    def _load_dump(self, dump_path: str):
        log.debug("[dump] path= {}", os.path.abspath(dump_path))

        self.sample_name = "DUMP"
        norm: dict = defaultdict(int)
        muts: dict = defaultdict(int)
        with gzip.open(dump_path, "rb") as fd:
            log.warn("Loading debug dump from {}", dump_path)
            l, h, i = (
                struct.calcsize("<l"),
                struct.calcsize("<h"),
                struct.calcsize("<i"),
            )
            cn_len = struct.unpack("<l", fd.read(l))[0]
            for _ in range(cn_len):
                i, v = struct.unpack("<ll", fd.read(l + l))
                self._dump[0][i] = v

            ld = struct.unpack("<l", fd.read(l))[0]
            for _ in range(ld):
                ref_start, ref_end, read_len, num_mutations = struct.unpack(
                    "<llhh", fd.read(l + l + h + h)
                )
                ref_end += ref_start
                for j in range(ref_start, ref_end):
                    norm[j] += 1
                mut_set = set()
                for _ in range(num_mutations):
                    mut_start, op_len = struct.unpack("<hh", fd.read(h + h))
                    mut_start += ref_start
                    op = fd.read(op_len).decode("ascii")
                    muts[mut_start, op] += 1
                    mut_set.add((mut_start, op))
                    if op[:3] == "del":
                        for j in range(0, len(op) - 3):
                            norm[mut_start + j] -= 1
                    elif op[:3] == "ins":
                        self._insertion_counts[mut_start, len(op) - 3] += 1
                    else:
                        norm[mut_start] -= 1
                mut_set_pos = {p for p, _ in mut_set}
                for pos, op in self._multi_sites.items():
                    if pos not in mut_set_pos:
                        continue
                    ll, r = op.split(">")
                    if all(
                        (pos + p, f"{ll[p]}>{r[p]}") in mut_set
                        for p in range(len(ll))
                        if ll[p] != "."
                    ):
                        for p in range(len(ll)):
                            if ll[p] != ".":
                                muts[pos + p, f"{ll[p]}>{r[p]}"] -= 1
                                if p:
                                    norm[pos + p] += 1
                        muts[pos, op] += 1
                for data in self._insertion_reads.values():
                    data[ref_start, ref_end, read_len] += 1
        return norm, muts

    def _make_coverage(self, gene, norm, muts, threshold):
        # Establish the coverage dictionary
        coverage: Dict[int, Dict[str, int]] = dict()
        for pos, cov in norm.items():
            if cov == 0:
                continue
            if pos not in coverage:
                coverage[pos] = {}
            coverage[pos]["_"] = cov
        bounds = min(gene.chr_to_ref), max(gene.chr_to_ref)
        for m, cov in muts.items():
            pos, mut = m
            if pos not in coverage:
                coverage[pos] = {}
            if not bounds[0] <= pos <= bounds[1] and mut[:3] != "ins":
                mut = "_"  # ignore mutations outside of the region of interest
            if mut not in coverage[pos]:
                coverage[pos][mut] = 0
            coverage[pos][mut] += cov
        self._group_indels(gene, coverage)
        for pos, op in self._multi_sites.items():
            if pos in coverage and op in coverage[pos]:
                log.debug(f"[sam] multi-SNP {pos}{op} with {coverage[pos][op]} reads")
        for mut in self._insertion_sites:
            if mut.pos in coverage and mut.op in coverage[mut.pos]:
                total = sum(v for o, v in coverage[mut.pos].items() if o[:3] != "ins")
                if total > 0:
                    coverage[mut.pos][mut.op] = self._correct_ins_coverage(mut, total)

        self.coverage = Coverage(
            {p: {m: v for m, v in coverage[p].items() if v > 0} for p in coverage},
            threshold,
            self._dump[0],
            self.sample_name,
        )

    def _group_indels(self, gene, coverage):
        # Group ambiguous deletions
        indels = defaultdict(set)
        for m in gene.mutations:
            if "del" in m[1]:
                indels[m[0]].add(m[1])
        for pos in coverage:
            for mut in coverage[pos]:
                if mut[:3] != "del" or "N" in mut[3:]:
                    continue
                potential = [pos]
                sz = len(mut[3:])
                deleted = gene[pos : pos + sz]
                for p in range(pos - sz, -1, -sz):
                    if gene[p : p + sz] != deleted:
                        break
                    potential.append(p)
                for p in range(pos + sz, max(coverage), sz):
                    if gene[p : p + sz] != deleted:
                        break
                    potential.append(p)
                potential = [
                    p
                    for p in set(potential) & set(indels)
                    for m in indels[p]
                    if m == mut
                ]
                if len(potential) == 1 and potential[0] != pos:
                    new_pos = potential[0]
                    log.debug(
                        f"[sam] relocate {coverage[pos][mut]} "
                        + f"from {pos+1}{mut} to {new_pos+1}{mut}"
                    )
                    if mut not in coverage[new_pos]:
                        coverage[new_pos][mut] = 0
                    coverage[new_pos][mut] += coverage[pos][mut]
                    coverage[pos][mut] = 0
        # Group ambiguous insertions
        indels = defaultdict(set)
        for m in gene.mutations:
            if "ins" in m[1]:
                indels[m[0]].add(m[1])
        for pos in coverage:
            for mut in coverage[pos]:
                if mut[:3] != "ins":
                    continue
                inserted, sz = mut[3:], len(mut[3:])
                potential = []
                for p in range(pos, -1, -sz):
                    potential.append(p)
                    if gene[p : p + sz] != inserted:
                        break
                for p in range(pos, max(coverage), sz):
                    potential.append(p)
                    if gene[p : p + sz] != inserted:
                        break
                potential = [
                    p
                    for p in set(potential) & set(indels)
                    for m in indels[p]
                    if m == mut
                ]
                if len(potential) == 1 and potential[0] != pos:
                    new_pos = potential[0]
                    log.debug(
                        f"[sam] relocate {coverage[pos][mut]} "
                        + f"from {pos+1}{mut} to {new_pos+1}{mut}"
                    )
                    if mut not in coverage[new_pos]:
                        coverage[new_pos][mut] = 0
                    coverage[new_pos][mut] += coverage[pos][mut]

                    L = len(mut) - 3
                    self._insertion_counts[new_pos, L] += self._insertion_counts[pos, L]
                    self._insertion_counts[pos, L] = 0

                    coverage[pos][mut] = 0

    def _parse_read(self, read, gene: Gene, norm, muts):
        """
        Parse a :obj:`pysam.AlignedSegment` read.

        :param read: pysam read.
        :param gene: Gene instance.
        :param norm: Positions within the read that have not been mutated.
        :param muts: Positions within the read that have been mutated.

        .. note:: `norm` and `muts` are modified.
        """

        if not _in_region(
            gene.get_wide_region(), read, self._prefix
        ):  # ensure that it is a proper gene read
            return None
        if not read.cigartuples:  # only valid alignments
            return None
        if read.flag & 0x800:  # avoid supplementary alignments
            return None
        if "H" in read.cigarstring:  # avoid hard-clipped reads
            return None

        dump_arr = []
        start, s_start = read.reference_start, 0
        for op, size in read.cigartuples:
            if op == 2:  # Deletion
                mut = (
                    start,
                    "del" + gene[start : start + size],
                )
                muts[mut] += 1
                dump_arr.append(mut)
                start += size
            elif op == 1:  # Insertion
                mut = (start, "ins" + read.query_sequence[s_start : s_start + size])
                muts[mut] += 1
                # HACK: just store the length due to seq. errors
                self._insertion_counts[start, size] += 1
                dump_arr.append(mut)
                s_start += size
            elif op == 4:  # Soft-clip
                s_start += size
            elif op in [0, 7, 8]:  # M, X and =
                for i in range(size):
                    if (
                        start + i in gene
                        and gene[start + i] != read.query_sequence[s_start + i]
                    ):
                        mut = (
                            start + i,
                            f"{gene[start + i]}>{read.query_sequence[s_start + i]}",
                        )
                        dump_arr.append(mut)
                        muts[mut] += 1
                    else:  # We ignore all mutations outside the RefSeq region
                        norm[start + i] += 1
                start += size
                s_start += size

        dump_arr_pos = {p for p, _ in dump_arr}
        for pos, op in self._multi_sites.items():
            if pos not in dump_arr_pos:
                continue
            l, r = op.split(">")
            if all(
                (pos + p, f"{l[p]}>{r[p]}") in dump_arr
                for p in range(len(l))
                if l[p] != "."
            ):
                for p in range(len(l)):
                    if l[p] != ".":
                        muts[pos + p, f"{l[p]}>{r[p]}"] -= 1
                        if p:
                            norm[pos + p] += 1
                muts[pos, op] += 1

        read_pos = (read.reference_start, start, len(read.query_sequence))
        for data in self._insertion_reads.values():
            data[read_pos] += 1  # type: ignore
        return read_pos, dump_arr

    def _correct_ins_coverage(self, mut: Mutation, total) -> int:
        """
        Fix low coverage of large tandem insertions.

        Targets the cases where reference looks like
            ...X...
        and the donor genome looks like
            ...XX... (i.e. X is a tandem insertion).

        Any read that covers only one tandem (e.g. read X...) will get perfectly aligned
        (without trigerring an insertion tag) to the tail of the tandem insertion.
        This results in under-estimation of the insertion abundance will (as aligner
        could not assign `I` CIGAR to the correct insertion).

        This function attempts to correct this bias.
        """

        MARGIN_RATIO = 0.20
        INSERT_LEN_RESCALE = 10.0
        INSERT_LEN_AMPLIFY = 3

        orig_cov = self._insertion_counts[mut.pos, len(mut.op) - 3]
        new_total = 0.0
        for (orig_start, orig_end, read_len), cov in self._insertion_reads[mut].items():
            ins_len = len(mut.op) - 3
            # HACK: This is horrible way of detecting *40 (min 20% on the sides? max??)
            min_margin = MARGIN_RATIO * max(1, ins_len / INSERT_LEN_RESCALE) * read_len
            if orig_end >= mut.pos + max(
                INSERT_LEN_AMPLIFY * ins_len, min_margin
            ) and orig_start <= mut.pos - max(
                (INSERT_LEN_AMPLIFY - 1) * ins_len, min_margin
            ):
                new_total += cov
        new_total += orig_cov
        new_cov = int(total * (orig_cov / new_total))
        if new_cov > orig_cov:
            log.debug(f"[sam] rescale {mut} from {orig_cov} to {new_cov}")
            return new_cov
        return orig_cov

    # ----------------------------------------------------------------------------------
    # Coverage-rescaling functions
    # ----------------------------------------------------------------------------------

    def detect_cn(self, gene: Gene, profile: str, cn_region: GRange) -> None:
        """
        Rescale the ``self.coverage`` to fit the sequencing profile.

        :param gene: Gene instance.
        :param profile: Profile identifier. Can be a SAM/BAM file as well.
        :param cn_region: Coordinates of the copy-number neutral region.

        .. notes:: This function assumes that ``self.coverage`` is set and modifies it.
        """

        if os.path.exists(profile) and os.path.isfile(profile):
            ext = os.path.splitext(profile)
            if ext[-1] in [".bam", ".sam"]:
                prof = self._load_profile(
                    profile,
                    is_bam=True,
                    gene=gene,
                    cn_region=cn_region,
                )
            else:
                prof = self._load_profile(profile)
        else:
            profile_path = script_path(
                "aldy.resources.profiles/{}.yml".format(profile.lower())
            )
            prof = self._load_profile(profile_path)
            if "neutral" not in prof or "cn" not in prof["neutral"]:
                raise AldyException("Profile missing neutral region")
            if gene.name not in prof:
                raise AldyException(f"Profile missing {gene.name}")
        self.coverage._normalize_coverage(
            prof[gene.name], gene.regions, cn_region, prof["neutral"]["cn"][0]
        )

    def _load_profile(
        self,
        profile_path: str,
        is_bam: bool = False,
        gene: Optional[Gene] = None,
        cn_region: Optional[GRange] = None,
    ) -> Dict[str, Dict[str, List[float]]]:
        """
        Load a coverage profile.

        :param profile_path:
            Path to a profile file.
            SAM/BAM files are also accepted if `is_bam` is set
            (profile will be dynamically calculated in that case).
        :param is_bam: A flag indicating if the `profile_path` is SAM/BAM or not.
        :param gene_region: Region to be extracted from the profile.
        :param cn_region: Copy-number neutral region to be extracted from the profile.

        :return: A profile dictionary where keys are chromosome IDs (e.g. '7' for chr7)
                 and values are dictionaries that map the genomic locus to the
                 corresponding profile coverage.
                 This is a :obj:`collections.defaultdict` that uses 0 as a placeholder
                 for the missing loci.

        :raise: :obj:`aldy.common.AldyException` if ``is_bam`` is set
                but ``gene_region`` and ``cn_region`` are not.
        """

        if is_bam:
            assert gene and cn_region
            regions = {}
            for gi, gr in enumerate(gene.regions):
                for r, rng in gr.items():
                    regions[gene.name, r, gi] = rng
            regions["neutral", "cn", 0] = cn_region
            return load_sam_profile(profile_path, regions=regions)
        else:
            with open(profile_path) as f:
                return yaml.safe_load(f)

    def _dump_alignments(self, debug: str):
        with gzip.open(f"{debug}.dump", "wb") as fd:
            cn, dump_data = self._dump
            fd.write(struct.pack("<l", len(cn)))
            for i, v in cn.items():
                fd.write(struct.pack("<ll", i, v))
            fd.write(struct.pack("<l", len(dump_data)))
            for (s, e, l), m in dump_data:
                fd.write(struct.pack("<llhh", s, e - s, l, len(m)))
                for p, md in m:
                    fd.write(struct.pack("<hh", p - s, len(md)))
                    fd.write(md.encode("ascii"))


def load_sam_profile(
    sam_path: str,
    factor: float = 2.0,
    regions: Dict[Tuple[str, str, int], GRange] = dict(),
    cn_region: Optional[GRange] = None,
) -> Dict[str, Dict[str, List[float]]]:
    """
    Load the profile information from a SAM/BAM file.

    Returns:
        list[str, str, int, float]: list of tuples
        ``(gene_name, chromosome, loci, coverage)``.

    Params:
        factor (float):
            Scaling factor. Default is 2.0 (for two copies).
        regions (list[:obj:`GRange`], optional):
            List of regions to be extracted.

    Notes:
        Profiles that were used in Aldy paper:

            1. PGRNseq-v1/v3: NA17642 was used for all genes
                (n.b. PGXT147 with rescale 2.52444127771 was used for CYP2B6 beta).
            2. PGRNseq-v2: NA19789.bam was used for all genes.
            3. Illumina: by definition contains all ones (uniform coverage profile).
    """

    if len(regions) == 0:
        import pkg_resources

        gene_regions = {}
        for g in sorted(pkg_resources.resource_listdir("aldy.resources", "genes")):
            if g[-4:] != ".yml":
                continue
            gg = Gene(script_path(f"aldy.resources.genes/{g}"))
            for gi, gr in enumerate(gg.regions):
                for r, rng in gr.items():
                    gene_regions[gg.name, r, gi] = rng
    else:
        gene_regions = regions
    gene_regions["neutral", "cn", 0] = (
        cn_region if cn_region else DEFAULT_CN_NEUTRAL_REGION["hg19"]
    )

    chr_regions: Dict[str, Tuple[int, int]] = {}
    for c, s, e in gene_regions.values():
        if c not in chr_regions:
            chr_regions[c] = (s, e)
        else:
            chr_regions[c] = (min(s, chr_regions[c][0]), max(e, chr_regions[c][1]))

    cov: dict = defaultdict(lambda: defaultdict(int))
    for c, (s, e) in natsorted(chr_regions.items()):
        with pysam.AlignmentFile(sam_path) as sam:
            region = GRange(c, s, e).samtools(
                pad_left=1000,
                pad_right=1000,
                prefix=_chr_prefix(c, [x["SN"] for x in sam.header["SQ"]]),
            )
            log.info("Scanning {}...", region)
            try:
                for read in sam.fetch(region=region):
                    start, s_start = read.reference_start, 0
                    if not read.cigartuples:
                        continue

                    for op, size in read.cigartuples:
                        if op == 2:
                            for i in range(size):
                                cov[c][start + i] += 1
                            start += size
                        elif op == 1:
                            s_start += size
                        elif op == 4:
                            s_start += size
                        elif op in [0, 7, 8]:
                            for i in range(size):
                                cov[c][start + i] += 1
                            start += size
                            s_start += size
            except ValueError:
                log.warn("Cannot fetch {}", region)

    d: Any = {}
    for (g, r, ri), (c, s, e) in gene_regions.items():
        if g not in d:
            d[g] = {}
        if r not in d[g]:
            d[g][r] = [0]
        if ri >= len(d[g][r]):
            d[g][r].append(0)
        d[g][r][ri] = sum(cov[c][i] * (factor / 2.0) for i in range(s, e))
    return d


def detect_genome(sam_path: str) -> Tuple[str, Optional[str]]:
    try:
        with pysam.AlignmentFile(sam_path) as sam:
            try:
                sam.check_index()
            except AttributeError:
                pass  # SAM files do not have an index. BAMs might also lack it
            except ValueError:
                raise AldyException(
                    f"File {sam_path} has no index (it must be indexed)"
                )

            #: str: Check if we need to append 'chr' or not
            _prefix = _chr_prefix("1", [x["SN"] for x in sam.header["SQ"]])
            # Detect the genome
            hg19_lens = {"1": 249250621, "10": 135534747, "22": 51304566}
            hg38_lens = {"1": 248956422, "10": 133797422, "22": 50818468}
            chrs = {x["SN"]: int(x["LN"]) for x in sam.header["SQ"]}
            is_hg19, is_hg38 = True, True
            for c in "1 10 22".split():
                is_hg19 &= chrs[_prefix + c] == hg19_lens[c]
                is_hg38 &= chrs[_prefix + c] == hg38_lens[c]
            if is_hg19:
                return "sam", "hg19"
            elif is_hg38:
                return "sam", "hg38"
            else:
                return "sam", None
    except (ValueError, OSError):
        try:
            with pysam.VariantFile(sam_path):
                return "vcf", None
        except (ValueError, OSError):
            pass
    return "", None


def _chr_prefix(ch: str, chrs: List[str]) -> str:  # assumes ch is not prefixed
    """
    Check if `ch` (*without any chr prefix*) should be prefixed with "chr" or not.

    Returns:
        str: Prefix to be prepended to chromosome (empty if needed).

    Params:
        ch (str): chromosome name
        chrs (list[str]): list of chromosome names in the alignment file
    """
    if ch not in chrs and "chr" + ch in chrs:
        return "chr"
    return ""


def _in_region(region: GRange, read: pysam.AlignedSegment, prefix: str) -> bool:
    """
    Returns ``True`` if a read is located within a given gene region.

    Notes:
        The region is padded with 500bp on the left side.
    """

    return (
        read.reference_id != -1
        and read.reference_name == prefix + region.chr
        and region.start - 500 <= read.reference_start <= region.end
    )


def load_phase(gene: Gene, path: str):
    """Loads a HapTree-X/HapCUT2 phase file."""

    haplotypes: Tuple[List[Mutation], List[Mutation]] = ([], [])
    phases: List[List[Mutation]] = []

    g_chr, g_s, g_e = gene.get_wide_region()
    with open(path) as hap:
        for li, line in enumerate(hap):
            if line[:5] == "BLOCK":
                if haplotypes[0]:
                    phases += list(haplotypes)
                haplotypes = [], []
            elif line[:5] == "*****":
                continue
            else:
                ls = line.strip().split("\t")
                if len(ls) < 7:
                    raise AldyException(
                        f"Invalid phasing line {li + 1} in {path} (less than 7 columns)"
                    )
                _, gt0_, gt1_, chr, pos_, al0, al1, *_ = ls
                gt0, gt1, pos = int(gt0_), int(gt1_), int(pos_) - 1
                if gt0 + gt1 != 1:
                    continue
                chr = chr[3:] if chr.startswith("chr") else chr
                if chr != g_chr:
                    continue
                if pos < g_s or pos >= g_e:
                    continue

                if (pos, f"{al0}>{al1}") not in gene.mutations:
                    if (pos, f"{al1}>{al0}") in gene.mutations:
                        log.warn(f"reorienting {pos} {al0} {al1}")
                        al1, al0 = al0, al1
                        gt0, gt1 = gt1, gt0
                haplotypes[0].append(Mutation(pos, f"{al0}>{al1}" if gt0 else "_"))
                haplotypes[1].append(Mutation(pos, f"{al0}>{al1}" if gt1 else "_"))
    if haplotypes[0]:
        phases += [haplotypes[0], haplotypes[1]]
    return phases



# 786
# Aldy source: coverage.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, Tuple, Callable, List

import collections

from .common import log, AldyException
from .gene import Mutation, GRange


class Coverage:
    """
    Data structure that maintains the coverage information for a given sample.
    """

    def __init__(
        self,
        coverage: Dict[int, Dict[str, int]],
        threshold: float,
        cnv_coverage: Dict[int, int],
        sample: str = "sample",
    ) -> None:
        """
        Coverage initialization.

        :param coverage:
            Coverage for each locus within a sample represented as a dictionary that
            maps mutation (or a reference position indicated by `_`) to the number of
            reads supporting that mutation.
            For example, ``coverage[10]['A>G'] = 2`` means that there are 2 reads that
            have G (instead of A) at the genomic locus 10.
        :param threshold:
            Threshold `t` used for filtering out low-quality mutations.
            Any mutation with the coverage less than `t`% is filtered out.
            Ranges from 0 to 1 (normalized percentage).
        :param cnv_coverage:
            Coverage of the copy-number neutral region of the sample: each genomic locus
            within that region points to the corresponsing read coverage.
            Used for coverage rescaling.
        :param sample: Sample name. Default: 'sample'.
        """
        self._coverage = coverage
        self._threshold = threshold
        self._cnv_coverage = cnv_coverage
        self.sample = sample

        self._region_coverage: Dict[Tuple[int, str], float] = {}

    def __getitem__(self, mut: Mutation) -> float:
        """ :return: Coverage of the mutation ``mut``. """
        return self.coverage(mut)

    def coverage(self, mut: Mutation) -> float:
        """ :return: Coverage of the mutation ``mut``. """
        if mut.op in self._coverage[mut.pos]:
            return self._coverage[mut.pos][mut.op]
        else:
            return 0

    def total(self, pos: int) -> float:
        """ :return: Total coverage at the locus ``pos``. """
        if pos not in self._coverage:
            return 0
        return float(sum(v for p, v in self._coverage[pos].items() if p[:3] != "ins"))

    def percentage(self, m: Mutation) -> float:
        """ :return: Coverage of the mutation ``mut`` as a percentage (0-100%). """
        total = self.total(m.pos)
        if total == 0:
            return 0
        return 100.0 * self.coverage(m) / total

    def single_copy(self, pos: int, cn_solution) -> float:
        """
        :param pos: Genomic locus,
        :param cn_solution: Copy-number solution.
        :return: Coverage of a single gene copy at the locus ``pos``.
        """
        if cn_solution.position_cn(pos) == 0:
            return 0
        return max(1, self.total(pos)) / cn_solution.position_cn(pos)

    def region_coverage(self, gene: int, region: str) -> float:
        """ :return: Average coverage of the region ``region`` in ``gene``. """
        return self._region_coverage[gene, region]

    def average_coverage(self) -> float:
        """ :return: Average coverage of the sample. """
        return sum(self.total(pos) for pos in self._coverage) / float(
            len(self._coverage) + 0.1
        )

    def filtered(
        self, filter_fn: Callable[[Mutation, float, float, float], bool]
    ):  # -> Coverage
        """
        :param filter_fn:
            Function that performs mutation filtering with the following arguments:

                1. mut (:obj:`aldy.gene.Mutation`): mutation to be filtered
                2. cov (float): coverage of the mutation
                3. total (float): total coverage of the mutation locus
                4. thres (float): filtering threshold

            ``filter_fn`` returns ``False`` if a mutation is filtered out.

        :return: Filtered coverage.
        """

        cov = collections.defaultdict(
            lambda: collections.defaultdict(int),
            {
                pos: collections.defaultdict(
                    int,
                    {
                        o: c
                        for o, c in pos_mut.items()
                        if filter_fn(
                            Mutation(pos, o),  # type: ignore
                            c,
                            self.total(pos),
                            self._threshold,
                        )
                    },
                )
                for pos, pos_mut in self._coverage.items()
            },
        )

        new_cov = Coverage(
            cov, self._threshold, self._cnv_coverage, self.sample  # type: ignore
        )
        new_cov._region_coverage = self._region_coverage
        return new_cov

    def diploid_avg_coverage(self) -> float:
        """ :return: Average coverage of the copy-number neutral region. """
        return float(sum(self._cnv_coverage.values())) / abs(
            self._cn_region.end - self._cn_region.start
        )

    def _normalize_coverage(
        self,
        profile: Dict[str, List[float]],
        gene_regions: List[Dict[str, GRange]],
        cn_region: GRange,
        profile_cn: float,
    ) -> None:
        """
        Normalize the sample coverage to match the profile coverage.

        :param profile:
            Profile coverage in the form `chromosome: (position -> coverage)`.
        :param gene_regions: List of genic regions for each gene.
        :param cn_region: Copy-number neutral region.
        """

        self._cn_region: GRange = cn_region  #: GRange: store the CN-neutral region
        sam_ref = sum(
            self._cnv_coverage[i] for i in range(cn_region.start, cn_region.end)
        )

        if sam_ref == 0:
            raise AldyException(
                f"CN-neutral region {cn_region} has no reads. "
                + "Double check your input file for CYP2D8 (are you using hg19?), "
                + "or pass an alternative CN-neutral region via -n parameter."
            )
        cn_ratio = float(profile_cn) / sam_ref
        if cn_ratio == 0:
            raise AldyException("Invalid CN-neutral region in the provided profile.")
        log.debug("[coverage] scale_ratio: {:.1f}", 1 / cn_ratio)

        self._region_coverage = {}
        for gene, gr in enumerate(gene_regions):
            for region, rng in gr.items():
                s = sum(self.total(i) for i in range(rng.start, rng.end))  # !IMPORTANT
                p = profile[region][gene]
                self._region_coverage[gene, region] = (
                    (cn_ratio * float(s) / p) if p != 0 else 0.0
                )

    @staticmethod
    def basic_filter(mut: Mutation, cov: float, total: float, thres: float) -> bool:
        """
        Basic filtering function.
        """
        return cov >= max(1, total * thres)

    @staticmethod
    def cn_filter(
        mut: Mutation, cov: float, total: float, thres: float, cn_solution
    ) -> bool:
        """
        Filtering function that takes into the account the copy number of the mutation.
        """
        cn = cn_solution.position_cn(mut.pos)
        total = total / cn if cn > 0 else total

        return mut.op == "_" or cov >= max(1, total * thres)



#!/usr/bin/env python
# 786
# Aldy source: __main__.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Optional, Any

import logbook.more
import argparse
import os
import sys
import platform
import yaml
import datetime
import tempfile
import pkg_resources
import traceback
import pytest

from . import common
from .common import log, script_path, AldyException, td, parse_cn_region
from .gene import Gene
from .cn import LEFT_FUSION_PENALTY
from .sam import load_sam_profile
from .genotype import genotype
from .query import query
from .version import __version__


def get_version():
    return "{} {}".format(
        platform.system() if platform.system() != "Darwin" else "macOS",
        platform.platform()[6:]
        if platform.system() == "Linux"
        else platform.mac_ver()[0]
        if platform.system() == "Darwin"
        else platform.platform(),
    )


def main(argv):
    """
    The main entry point.
    """

    parser, args = _get_args(argv)

    # Set the logging verbosity
    level = (args.verbosity if "verbosity" in args else "I").lower()
    level = next(
        v
        for k, v in logbook.base._reverse_level_names.items()
        if k.lower().startswith(level)
    )

    # Set the command-line logging
    sh = logbook.more.ColorizedStderrHandler(
        format_string="{record.message}", level=level
    )
    sh.push_application()

    log.info(
        "🐿  Aldy v{} (Python {} on {})",
        __version__,
        platform.python_version(),
        get_version(),
    )
    log.info(
        "   (c) 2016-{} Aldy Authors. All rights reserved.\n"
        + "   Free for non-commercial/academic use only.",
        datetime.datetime.now().year,
    )

    try:
        if not args.subparser or args.subparser == "help":
            parser.print_help()
        elif args.subparser == "license":
            _print_licence()
        elif args.subparser == "test":
            _run_test()
        elif args.subparser in ["query", "q"]:
            q = args.gene
            if "*" in q:
                gene, q = q.split("*", maxsplit=1)
            else:
                gene, q = q, ""
            db_file = script_path("aldy.resources.genes/{}.yml".format(gene.lower()))
            if os.path.exists(db_file):
                gene_db = db_file
            else:
                gene_db = gene
            with open(gene_db):  # Check if file exists
                pass
            query(Gene(gene_db), q)
        elif args.subparser == "profile":
            p = load_sam_profile(
                args.file, cn_region=parse_cn_region(args.cn_neutral_region)
            )
            print(yaml.dump(p, default_flow_style=None))
        elif args.subparser in ["genotype", "g"]:
            # Prepare the list of available genes
            if args.gene.lower() == "all":
                avail_genes = pkg_resources.resource_listdir("aldy.resources", "genes")
                avail_genes = [
                    i[:-4] for i in avail_genes if len(i) > 4 and i[-4:] == ".yml"
                ]
                avail_genes = sorted(avail_genes)
            else:
                avail_genes = args.gene.lower().split(",")

            # Prepare the output file
            output = args.output
            if output == "-":
                output = sys.stdout
            elif output:
                output = open(output, "w")
            for gene in avail_genes:
                _genotype(gene, output, args)
            if output and output != sys.stdout:
                output.close()
        else:
            raise AldyException("Invalid sub-command " + args.subparser)
    except IOError as ex:
        if ex.filename is not None:
            log.critical("ERROR: {} cannot be accessed", ex.filename)
        else:
            log.critical("ERROR: {} cannot be accessed", str(ex))
        log.debug(ex)
        log.debug(traceback.format_exc())
        exit(1)
    except SystemExit as ex:
        log.debug(repr(ex))
        log.debug(traceback.format_exc())
        exit(ex.code)
    except Exception as ex:
        log.critical(
            f"ERROR: gene= {args.gene}, profile= {args.profile}, file= {args.file}"
        )
        log.critical(repr(ex))
        log.warn(traceback.format_exc())
        exit(1)
    except:  # noqa
        exc = sys.exc_info()[0]
        log.critical(
            f"ERROR: gene= {args.gene}, profile= {args.profile}, file= {args.file}"
        )
        log.critical("Unrecoverable error: {}", repr(exc))
        log.warn(traceback.format_exc())
        exit(1)


def _get_args(argv):
    """
    Parse command-line arguments.
    """

    parser = argparse.ArgumentParser(
        prog="aldy",
        description="Allelic decomposition and exact genotyping of highly polymorphic "
        + "and structurally variant genes",
    )

    base = argparse.ArgumentParser(add_help=False)
    base.add_argument(
        "--verbosity",
        "-v",
        default="INFO",
        help=td(
            """Logging verbosity:
               - T (trace)
               - D (debug)
               - I (info) and
               - W (warn).
               Default is "I" (info)."""
        ),
    )

    subparsers = parser.add_subparsers(dest="subparser")

    genotype_parser = subparsers.add_parser(
        "genotype",
        help="Call the most likely genotype and diplotype within a sample.",
        parents=[base],
    )
    genotype_parser.add_argument(
        "file", nargs="?", help="Input file in SAM, BAM or CRAM format."
    )
    genotype_parser.add_argument(
        "--gene",
        "-g",
        default="all",
        help='Gene whose genotype is to be called. Default is "all" which calls '
        + "genotypes for all supported genes.",
    )
    genotype_parser.add_argument(
        "--profile",
        "-p",
        required=True,
        help=td(
            """Sequencing profile. The following profiles are supported:
               - illumina
               - wgs (same as illumina)
               - pgrnseq-v1
               - pgrnseq-v2,
               - pgrnseq-v3, and
               - exome.
               You can also provide a custom SAM/BAM file as a profile.
               Please check documentation for more details."""
        ),
    )
    genotype_parser.add_argument(
        "--threshold",
        "-T",
        default=50,
        help="Cut-off rate for variations (percent per copy). Default is 50.",
    )
    genotype_parser.add_argument(
        "--reference",
        "-r",
        default=None,
        help="Genome reference used for reading CRAM files",
    )
    genotype_parser.add_argument(
        "--cn-neutral-region",
        "-n",
        help=td(
            """Copy-number neutral region in the format chromosome:start-end
               (e.g. chr1:10000-20000).
               Default is CYP2D8 region within hg19 (22:42547463-42548249)."""
        ),
    )
    genotype_parser.add_argument(
        "--output",
        "-o",
        default=None,
        help="Output file location. Default is [input].[gene].aldy.",
    )
    genotype_parser.add_argument(
        "--solver",
        "-s",
        default="any",
        help=td(
            """ILP solver:
               - gurobi (Gurobi)
               - scip (SCIP)
               - cbc (Google OR-Tools/CBC)
               - any (attempts to use Gurobi, then SCIP, then CBC).
               Default is "any"."""
        ),
    )
    genotype_parser.add_argument(
        "--gap",
        "-G",
        default=0,
        help=td(
            """Solver optimality gap.
               Any solution whose score is less than (1+gap) times the optimal solution
               score will be reported.
               Default is 0 (report only optimal solutions)."""
        ),
    )
    # genotype_parser.add_argument('--remap', default=0,
    #   help='Realign reads for better mutation calling.
    # Requires samtools and bowtie2 in $PATH.')
    genotype_parser.add_argument(
        "--debug",
        default=None,
        help="Create a directory that will contain the debug information "
        + "and core dumps.",
    )
    genotype_parser.add_argument("--log", "-l", default=None, help="Log file location")
    genotype_parser.add_argument(
        "--fusion-penalty",
        "-f",
        default=LEFT_FUSION_PENALTY,
        help="Fusion penalty. Use higher values to avoid fusions. "
        + f"Default is {LEFT_FUSION_PENALTY}.",
    )
    genotype_parser.add_argument(
        "--max-minor-solutions",
        default=1,
        help="Maximum number of minor solutions to report for each major solution. "
        + "Default is 1.",
    )
    genotype_parser.add_argument(
        "--cn",
        "-c",
        default=None,
        help=td(
            """Manually set the copy number configuration as a list of comma-separated
               configuration IDs (e.g. "CN1,CN2").
               For a list of supported configuration IDs, please run `aldy show`.
               For the most common diploid case that contains no structural variations
               (e.g. two copies of the main gene), use 1,1."""
        ),
    )
    genotype_parser.add_argument(
        "--multiple-warn-level",
        "-W",
        default=1,
        help="Show warning if multiple solutions are found. "
        + "Can be 1 (warn after genotyping) or 2 "
        + "(also warn if there are multiple major solutions)."
        + "Default is 1 (warn after the genotyping).",
    )
    genotype_parser.add_argument("--phase", help="Use phase file.")

    _ = subparsers.add_parser(
        "test",
        parents=[base],
        help="Run Aldy test suite. Recommended prior to the first use",
    )

    _ = subparsers.add_parser("license", parents=[base], help="Show Aldy license")

    show_parser = subparsers.add_parser(
        "query",
        aliases=["q"],
        parents=[base],
        help="Query database definitions for a given gene.",
    )
    show_parser.add_argument("gene", help="Gene or allele to show.")

    profile_parser = subparsers.add_parser(
        "profile",
        parents=[base],
        help=td(
            """Generate a sequencing profile for a SAM/BAM/CRAM file.
               Please check the documentation for more details."""
        ),
    )
    profile_parser.add_argument("file", nargs="?", help="SAM/BAM/CRAM file.")
    profile_parser.add_argument(
        "--cn-neutral-region",
        "-n",
        default=None,
        help=td(
            """Copy-number neutral region in the format chromosome:start-end
               (e.g. chr1:10000-20000).
               Default is CYP2D8 region within hg19 (22:42547463-42548249)."""
        ),
    )

    _ = subparsers.add_parser(
        "help", parents=[base], help="Show program usage and exit."
    )

    return parser, parser.parse_args(argv)


def _print_licence():
    """
    Print Aldy license.
    """
    with open(script_path("aldy.resources/LICENSE.rst")) as f:
        for ll in f:
            print(ll.strip())


def _genotype(gene: str, output: Optional[Any], args) -> None:
    """
    Genotype a file.

    Args:
        gene (str)
        output (str, optional)
        args: remaining command-line arguments

    Raises:
        :obj:`aldy.common.AldyException` if ``cn_region`` is invalid.
    """

    cn_region = parse_cn_region(args.cn_neutral_region)
    cn_solution = args.cn
    if cn_solution:
        cn_solution = cn_solution.split(",")

    threshold = float(args.threshold) / 100

    def run(debug):
        log.trace(
            "\n[main] arguments= {}",
            " ".join(k + "=" + str(v) for k, v in vars(args).items() if k is not None),
        )
        log.info("Genotyping sample {}...", os.path.basename(args.file))
        try:
            _ = genotype(
                gene_db=gene,
                sam_path=args.file,
                profile=args.profile,
                output_file=output,
                cn_region=cn_region,
                cn_solution=cn_solution,
                threshold=threshold,
                solver=args.solver,
                fusion_penalty=float(args.fusion_penalty),
                reference=args.reference,
                gap=float(args.gap),
                max_minor_solutions=int(args.max_minor_solutions),
                debug=debug,
                multiple_warn_level=args.multiple_warn_level,
                phase=args.phase,
                report=True,
            )
        except AldyException as ex:
            log.critical(
                f"ERROR: gene= {gene}, profile= {args.profile}, file= {args.file}"
            )
            log.error(str(ex))

    if args.log:
        fh = logbook.FileHandler(args.log, mode="w", bubble=True, level="DEBUG")
        fh.formatter = lambda record, _: record.message
        fh.push_application()
    if args.debug:
        with tempfile.TemporaryDirectory() as tmp:
            prefix = None
            try:
                prefix = f"{tmp}/{os.path.splitext(os.path.basename(args.file))[0]}"
                log_output = f"{prefix}.log"
                fh = logbook.FileHandler(
                    log_output, mode="w", bubble=True, level="TRACE"
                )
                fh.formatter = lambda record, _: "[{}:{}/{}] {}".format(
                    record.level_name[0],
                    os.path.splitext(os.path.basename(record.filename))[0],
                    record.func_name,
                    record.message,
                )
                fh.push_application()
                log.debug(f"Using {tmp} as temporary debug directory")
                run(prefix)
            finally:
                if prefix:
                    log.info("Preparing debug archive...")
                    with open(f"{prefix}.yml", "w") as f:
                        yaml.Dumper.ignore_aliases = lambda *args: True  # type: ignore
                        yaml.dump(common.json, f, default_flow_style=None)
                    os.system(f"tar czf {args.debug}.tar.gz -C {tmp} .")
    else:
        run(None)


def _run_test() -> None:
    """
    Run the Aldy test suite.
    """

    pytest.main(["--pyargs", "aldy"])


def console():
    main(sys.argv[1:])


if __name__ == "__main__":
    # import cProfile
    # cProfile.run("console()")
    console()



# 786
# Aldy source: genotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import List, Optional, Any, Set

import os
import sys

from . import sam
from . import cn
from . import major
from . import minor
from . import diplotype
from . import solutions

from .common import (
    colorize,
    log,
    script_path,
    json,
    AldyException,
    SOLUTION_PRECISION,
)
from .gene import Gene, GRange
from .diplotype import OUTPUT_COLS
from .lpinterface import model as lp_model


def genotype(
    gene_db: str,
    sam_path: str,
    profile: str,
    output_file: Optional[Any] = sys.stdout,
    cn_region: Optional[GRange] = None,
    cn_solution: Optional[List[str]] = None,
    threshold: float = 0.5,
    fusion_penalty: float = cn.LEFT_FUSION_PENALTY,
    solver: str = "any",
    reference: Optional[str] = None,
    gap: int = 0,
    max_minor_solutions: int = 1,
    debug: Optional[str] = None,
    multiple_warn_level: int = 1,
    phase: Optional[str] = None,
    report: bool = False,
    genome=None,
) -> List[solutions.MinorSolution]:
    """
    Genotype a sample.

    Returns:
        list[:obj:`aldy.solutions.MinorSolution`]: List of genotype solutions.

    Args:
        gene_db (str):
            Gene name (if it is located in the Aldy's gene database)
            or the location of a gene database in YML format.
        sam_path (str):
            Location of SAM/BAM/CRAM file that is to be analyzed.
        profile (str):
            Coverage profile (e.g. 'illumina').
            Can be ``None`` if ``cn_solution`` is provided.
        cn_region (:obj:`aldy.common.GRange`, optional):
            Copy-number neutral region.
            Can be ``None`` (will use the default CYP2D8 region or ``None``
            if ``cn_solution`` is provided).
        output_file (file, optional):
            Location of an output decomposition file.
            Provide ``None`` for no output.
            Default is ``sys.stdout``.
        cn_solution (list[str], optional):
            User-specified list of the copy number configurations.
            Copy-number detection solver will not run is this parameter is set.
            Default is ``None``.
        threshold (float):
            Mutation filtering threshold.
            Default is 0.5 (for 50%).
        fusion_penalty (float):
            Fusion penalty. Use higher values to avoid fusions.
            Default is 0.1.
        solver (str):
            ILP solver. Check :obj:`aldy.lpinterface` for the list of available solvers.
            Default is ``'any'``.
        phase (str, optional):
            Location of HapCUT or HapTree-compatible phased blocks.
        reference (str, optional):
            A reference genome for reading CRAM files.
            Default is ``None``.
        gap (float):
            Relative optimality gap. Use non-zero values to allow non-optimal solutions.
            Default is 0 (reports only optimal solutions).
        max_minor_solutions (int):
            Maximum number of minor solutions to report for each major solution.
            Default is 1.
        debug (str, optional):
            Prefix for debug information and core dump files.
            ``None`` for no debug information.
            Default is ``None``.

    Raises:
        :obj:`aldy.common.AldyException` if the average coverage is too low (below 2).
    """

    # Test the existence of LP solver
    _ = lp_model("init", solver)

    sample_name = os.path.splitext(os.path.basename(sam_path))[0]
    with open(sam_path):  # Check if file exists
        pass
    kind, g = sam.detect_genome(sam_path)
    # assert kind in ["vcf", "sam", "dump"]
    if genome is None:
        genome = g
        if not genome:
            log.warn("WARNING: Cannot detect genome, defaulting to hg19.")
            genome = "hg19"
        log.debug(f"[genotype] reference= {genome}")
    if genome not in ["hg19", "hg38"]:
        raise AldyException(f"Unknown genome {genome}")

    # Load the gene specification
    db_file = script_path("aldy.resources.genes/{}.yml".format(gene_db.lower()))
    if os.path.exists(db_file):
        gene_db = db_file
    with open(gene_db):  # Check if file exists
        pass
    gene = Gene(gene_db, genome=genome)

    if not cn_region:
        cn_region = sam.DEFAULT_CN_NEUTRAL_REGION[genome]
    if profile in ["exome", "wxs"]:
        log.warn("WARNING: Copy-number calling is not available for exome data.")
        log.warn(
            "WARNING: Aldy will NOT be able to detect gene duplications, "
            + "deletions and fusions."
        )
        log.warn(
            "WARNING: Calling of alleles that are defined by non-exonic mutations "
            + "is not available."
        )
        log.warn("         Results might not be biologically relevant!")
        cn_region = None
        cn_solution = ["1", "1"]
        profile = "illumina"
    elif profile == "wgs":
        profile = "illumina"
    elif profile == "pgrnseq-v1":
        profile = "pgx1"
    elif profile == "pgrnseq-v2":
        profile = "pgx2"
    elif profile == "pgrnseq-v3":
        profile = "pgx3"

    if kind == "vcf":
        log.warn("WARNING: Using VCF file. Copy-number calling is not available.")
        cn_region = None
        cn_solution = ["1", "1"]
        sample = sam.Sample(gene=gene, vcf_path=sam_path, debug=debug)
    else:
        sample = sam.Sample(
            gene=gene,
            sam_path=sam_path,
            threshold=threshold,
            profile=profile,
            reference=reference,
            cn_region=None if cn_solution else cn_region,
            debug=debug,
        )
        avg_cov = sample.coverage.average_coverage()
        if avg_cov < 2:
            raise AldyException(
                f"Average coverage of {avg_cov:.2f} for gene {gene.name} is too low; "
                + f"skipping gene {gene.name}. "
                + f"Please ensure that {gene.name} is present in the input SAM/BAM."
            )
        elif avg_cov < 20:
            log.warn(
                f"Average sample coverage is {avg_cov}. "
                + "We recommend at least 20x coverage for the best results."
            )

    json.update({"sample": os.path.basename(sam_path).split(".")[0], "gene": gene.name})

    # Get copy-number solutions
    cn_sols = cn.estimate_cn(
        gene,
        sample.coverage,
        solver=solver,
        gap=gap,
        user_solution=cn_solution,
        fusion_penalty=fusion_penalty,
        debug=debug,
    )

    # Add SLACK to each score to avoid division by zero or other numerical issues
    # when the optimum is close to zero.
    # HACK: Value of 1 (1 gene copy) seems to be reasonable, but is not
    # theoretically guaranteed to be better than any other value.
    SLACK = 1

    # Get major solutions and pick the best one
    if multiple_warn_level >= 3 and len(cn_sols) > 1:
        log.warn("WARNING: multiple gene structures found!")
    log.info(f"Potential {gene.name} gene structures for {sample_name}:")
    major_sols: list = []
    cn_sols = sorted(cn_sols, key=lambda m: (int(1000 * m.score), m._solution_nice()))
    if len(cn_sols) == 0:
        raise AldyException("No solutions found!")
    min_cn_score = min(cn_sols, key=lambda m: m.score).score
    for i, cn_sol in enumerate(cn_sols):
        conf = 100 * (min_cn_score + SLACK) / (cn_sol.score + SLACK)
        log.info(f"  {i + 1:2}: {cn_sol._solution_nice()} (confidence: {conf:.0f}%)")
    log.debug("*" * 80)

    phases = sam.load_phase(gene, phase) if phase else None

    for i, cn_sol in enumerate(cn_sols):
        major_sols += major.estimate_major(
            gene,
            sample.coverage,
            cn_sol,
            solver=solver,
            gap=gap,
            identifier=i,
            debug=debug,
        )
    if len(major_sols) == 0:
        raise AldyException("No major solutions found!")

    major_sols = [
        solutions.MajorSolution(
            m.score * ((m.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.cn_solution,
            m.added,
        )
        for m in major_sols
    ]
    min_major_score = min(major_sols, key=lambda m: m.score).score
    major_sols = sorted(
        [m for m in major_sols if m.score - min_major_score - gap < SOLUTION_PRECISION],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )

    if multiple_warn_level >= 2 and len(major_sols) > 1:
        log.warn("WARNING: multiple major solutions found!")
    log.info(f"Potential major {gene.name} star-alleles for {sample_name}:")
    for i, major_sol in enumerate(major_sols):
        conf = 100 * (min_major_score + SLACK) / (major_sol.score + SLACK)
        log.info(f"  {i + 1:2}: {major_sol._solution_nice()} (confidence: {conf:.0f}%)")
    log.debug("*" * 80)

    minor_sols = []
    for m in minor.estimate_minor(
        gene,
        sample.coverage,
        major_sols,
        solver,
        max_solutions=max_minor_solutions,
        debug=debug,
        phases=phases,
    ):
        n = solutions.MinorSolution(
            m.score
            * ((m.major_solution.cn_solution.score + SLACK) / (min_cn_score + SLACK)),
            m.solution,
            m.major_solution,
        )
        n.set_diplotype(m.get_diplotype())
        minor_sols.append(n)

    if len(minor_sols) == 0:
        raise AldyException(
            "Aldy could not phase any major solution.\n"
            + "Possible solutions:\n"
            + " - Check the coverage. Extremely low coverage prevents Aldy from "
            + "calling star-alleles.\n"
            + " - Run with --debug parameter and notify the authors of Aldy."
        )
    min_minor_score = min(minor_sols, key=lambda m: m.score).score
    minor_sols = sorted(
        [m for m in minor_sols if m.score - min_minor_score - gap < SOLUTION_PRECISION],
        key=lambda m: (int(1000 * m.score), m._solution_nice()),
    )
    log.debug("*" * 80)

    if multiple_warn_level >= 1 and len(minor_sols) > 1:
        log.warn("WARNING: multiple optimal solutions found!")
    log.info(
        f"{{}} {gene.name} star-alleles for {sample_name}:",
        "Best" if len(minor_sols) == 1 else "Potential",
    )
    is_vcf = output_file and output_file.name[-4:] == ".vcf"
    if output_file and not is_vcf:
        print("#" + "\t".join(OUTPUT_COLS), file=output_file)
    for i, minor_sol in enumerate(minor_sols):
        conf = 100 * (min_minor_score + SLACK) / (minor_sol.score + SLACK)
        log.info(
            f"  {i + 1:2}: {minor_sol.get_major_diplotype()}"
            + f" (confidence={conf:.0f}%)"
        )
        log.info(f"      Minor alleles: {minor_sol._solution_nice()}")
        if output_file and not is_vcf:
            print(f"#Solution {i + 1}: {minor_sol._solution_nice()}", file=output_file)
            diplotype.write_decomposition(
                sample_name, gene, i + 1, minor_sol, output_file
            )
        print(
            sample.coverage.sample,
            gene.name,
            minor_sol.get_major_diplotype().replace(" ", ""),
        )
    if is_vcf:
        diplotype.write_vcf(sample_name, gene, sample.coverage, minor_sols, output_file)

    if report:
        log.info(colorize(f"{gene.name} results:"))
        reported: Set[str] = set()
        for r in minor_sols:
            s = f"  - {r.get_major_diplotype()}"
            if s not in reported:
                log.info(colorize(s))
                log.info(f"    Minor: {r.get_minor_diplotype()}")
                log.info(f"    Legacy notation: {r.get_minor_diplotype(legacy=True)}")
                reported.add(s)
    if any(len(m.major_solution.added) > 0 for m in minor_sols) and not phase:
        novels = {
            gene.get_rsid(mm) for m in minor_sols for mm in m.major_solution.added
        }
        log.warn(
            "WARNING: mutations {} suggest presence of a novel major star-allele."
            + "\nHowever, such alleles cannot be determined without phasing data."
            + "\nPlease provide --phase parameter for Aldy to accurately call novel "
            + "major star-alleles.\nThe above-reported assignments of these mutations "
            + "are random.",
            ", ".join(sorted(novels)),
        )

    return minor_sols



# 786

# Aldy source: genotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.

from __future__ import print_function
from __future__ import division
from builtins import str

import os
import re
import sys
import logbook
import logbook.more
import collections
import platform
import textwrap
import pysam
import subprocess
import tempfile
import shutil

from . import cn
from . import gene
from . import sam
from . import lpinterface
from . import diplotype

# from . import remap_farid

from .common import *


def cmd(cmd):
    log.debug("Executing " + cmd)
    return subprocess.check_output(cmd, shell=True)


@timing
def optimize(gene, reads, alleles, copy_number):
    model = lpinterface.model("reads", "gurobi")

    # Binary variables for reads
    x = collections.defaultdict(dict)
    d = dict()
    for r in reads:
        for a in reads[r]:
            x[r][a] = model.addVar(vtype="B", name="x_{}_{}".format(a, r), update=False)
        d[r] = model.addVar(vtype="B", name="d_{}".format(r), update=False)

    # Coverage constraints
    z = collections.defaultdict(int)
    for r in reads:
        for a, (read, read_mut) in reads[r].items():
            for k in range(read.reference_start, read.reference_end):
                z[a, k] += x[r][a]
    y = dict()
    for (a, k), lhs in z.items():
        pos = alleles[a][0][0] + k

        v = model.addVar(lb=-model.INF, name="z_{}_{}".format(a, k), update=False)
        model.addConstr(
            v
            == lhs - copy_number[gene.region_at[pos]] * round(copy_number.baseline[pos])
        )

        y[a, k] = model.addVar(lb=0, name="y_{}_{}".format(a, k), update=False)
        model.model.addGenConstrAbs(y[a, k], v)

    ## Alternative: use mutation data
    ## Much slower model!
    # mut = collections.defaultdict(dict)
    # for r in reads:
    #    for a, (read, read_mut) in reads[r].items():
    #       for k in range(read.reference_start, read.reference_end):
    #          m = read_mut.get(k, '_')
    #          if m not in mut[a, k]:
    #             mut[a, k][m] = model.addVar(vtype='I', name='mut_{}_{}_{}'.format(a, k, m), update=False)
    #          z[a, k, m] += x[r][a]
    # for (a, k, m), lhs in z.items():
    #    pos = alleles[a][0][0] + k
    #    v = model.addVar(lb=-model.INF, name='z_{}_{}_{}'.format(a, k, m), update=False)
    #    model.addConstr(v == lhs - mut[a, k][m] * round(copy_number.baseline[pos]))
    #    y[a, k, m] = model.addVar(lb=0, name='y_{}_{}_{}'.format(a, k, m), update=False)
    #    model.model.addGenConstrAbs(y[a, k, m], v)
    # for a, k in mut:
    #    pos = alleles[a][0][0] + k
    #    model.addConstr(copy_number[gene.region_at[pos]] == model.quicksum(mut[a, k].values()))

    model.update()

    # Objective
    score = lambda r, m: len(r.query_sequence) - r.get_tag("NM")
    obj = model.quicksum(score(*reads[r][a]) * x[r][a] for r in x for a in x[r])
    obj -= model.quicksum(y.values())

    # Existence constraints
    for r in reads:
        model.addConstr(model.quicksum(x[r].values()) + d[r] == 1)

    # Solve
    def params(m):
        m.params.timeLimit = 30 * 60
        m.params.outputFlag = 1

    model.solve(objective=obj, method="max", init=params)

    # Remove bad reads
    result = {}
    deleted = 0
    for r, xa in x.items():
        if model.getValue(d[r]):
            deleted += 1
            continue
        correct_allele = [a for a in xa if model.getValue(xa[a])]
        assert len(correct_allele) == 1
        correct_allele = correct_allele[0]
        result[r] = reads[r][correct_allele]
    log.warn("Discarded {} reads".format(deleted))
    return result


def write_reads(sam_path, gene, alleles, reads, out_path):
    y = list(reads.values())
    y = sorted(
        y,
        key=lambda r: (
            alleles[r[0].reference_name][0][0] + r[0].reference_start,
            r[0].query_name,
        ),
    )

    with pysam.AlignmentFile(sam_path) as sam:
        with pysam.AlignmentFile(out_path, "wb", template=sam) as out:
            for x, _ in y:
                a = pysam.AlignedSegment()
                a.query_name = x.query_name.split("/")[0]
                a.query_sequence = x.query_sequence
                a.flag = x.flag
                a.reference_id = sam.get_tid("chr22")
                a.reference_start = x.reference_start + alleles[x.reference_name][0][0]
                a.mapping_quality = x.mapping_quality
                a.cigar = x.cigar
                a.next_reference_start = x.next_reference_start
                a.template_length = x.template_length
                a.query_qualities = x.query_qualities
                a.tags = x.tags
                out.write(a)
            cnv_chromosome, cnv_start, cnv_end = gene.cnv_region
            region = "chr{}:{}-{}".format(cnv_chromosome, cnv_start - 500, cnv_end + 1)
            for read in sam.fetch(region=region):
                out.write(read)
    cmd("samtools index {}".format(out_path))


def get_alleles(gene):
    def sequence(gene, id, pad=0):
        regs = [b for a, b in gene.regions.items() if a.startswith(str(id) + ".")]
        mi, ma = min(r[0] for r in regs), max(r[1] for r in regs)
        st = gene.region[1]
        return (
            (mi, ma),
            gene.seq[max(0, mi - st - pad) : min(ma - st + pad, len(gene.seq))],
        )

    return {"6": sequence(gene, "6"), "7": sequence(gene, "7")}


def realign(alleles, tempdir, sam_path):
    def makeref():
        for i in "67":
            with open("{}/{}.fa".format(tempdir, i), "w") as f:
                print(">{}".format(i), file=f)
                print("\n".join(textwrap.wrap(alleles[i][1])), file=f)
            out = cmd("bowtie2-build {0}/{1}.fa {0}/{1}".format(tempdir, i))
            log.trace(out)

    log.warn("Creating bowtie2 reference")
    makeref()

    def realign(path, regions):
        newpath = "{}/{}.fq".format(tempdir, os.path.basename(path))
        cmd(
            "samtools view -h {} {} | samtools fastq - > {}".format(
                path, " ".join(regions), newpath
            )
        )
        for i in "67":
            out = cmd(
                "bowtie2 -x {0}/{1} {2} -S {2}.{1}.sam".format(tempdir, i, newpath)
            )
            log.debug(out)

    log.warn("Aligning reads via bowtie2")
    regions = ["chr22:{}-{}".format(*alleles[i][0]) for i in alleles]
    realign(sam_path, regions)


def get_reads(alleles, tempdir, sam_path):
    reads = collections.defaultdict(dict)  # read -> allele
    for ref_id in "67":
        ref = alleles[ref_id][1]
        with pysam.AlignmentFile(
            "{}/{}.fq.{}.sam".format(tempdir, os.path.basename(sam_path), ref_id)
        ) as sam:
            for read in sam.fetch():
                start = read.reference_start
                start, s_start = read.reference_start, 0
                if not read.cigartuples or read.reference_name != ref_id:
                    continue

                seq = read.query_sequence
                rname = read.query_name
                muts = dict()

                for op, size in read.cigartuples:
                    if op == 2:
                        muts[start] = "DEL.{}".format(ref[start : start + size])
                        # muts.append(mut)
                        start += size
                    elif op == 1:
                        muts[start] = "INS.{}".format(
                            seq[s_start : s_start + size].lower()
                        )
                        # muts.append(mut)
                        s_start += size
                    elif op == 4:
                        s_start += size
                    elif op in [0, 7, 8]:
                        for i in range(size):
                            if (
                                0 <= start + i < len(ref)
                                and ref[start + i] != seq[s_start + i]
                            ):
                                muts[start + i] = "SNP.{}{}".format(
                                    ref[start + i], seq[s_start + i]
                                )
                                # muts.append(mut)
                        start += size
                        s_start += size
                reads[rname][ref_id] = (read, muts)
    return reads


@timing
def remap(
    sam_path, gene, sam, cn_sol, tempdir=None, force=True, cleanup=False, remap_mode=1
):
    if tempdir is None:
        tempdir = tempfile.mkdtemp(suffix="_tmpaldy", dir=".")
    log.critical("Temp is " + tempdir)

    if str(remap_mode) == "2":
        out = remap_farid.remap(sam_path, gene, sam, cn_sol, tempdir)
    else:
        log.critical("My remapper")

        alleles = get_alleles(gene)
        realign(alleles, tempdir, sam_path)

        log.warn("Reading realinged data...")
        reads = get_reads(alleles, tempdir, sam_path)

        log.warn("Optimizing...")
        reads = optimize(gene, reads, alleles, cn_sol)

        out = "{}.remap_ibrahim.bam".format(os.path.basename(sam_path))
        if os.path.exists(out) and not force:
            raise AldyException(
                "{} already exists--- make sure to use --force!".format(out)
            )
        write_reads(sam_path, gene, alleles, reads, out)

    if cleanup:
        shutil.rmtree(tempdir)
    return out



# 786
# Aldy source: cn.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Dict, List, Tuple, Optional

import copy
import re
from natsort import natsorted

from . import lpinterface

from .common import log, json, sorted_tuple, AldyException
from .gene import CNConfig, CNConfigType, Gene
from .coverage import Coverage
from .solutions import CNSolution


MAX_CN = 20.0
"""Maximum allowed copy number."""

# Model parameters

LEFT_FUSION_PENALTY = 0.1
"""Penalty for left fusions to account for their rarity (0 for no penalty)."""

PCE_PENALTY_COEFF = 1.5
"""Error penalty applied to the PCE region (1 for no penalty)."""

MAX_CN_ERROR = MAX_CN
"""Upper bound for absolute error."""

PARSIMONY_PENALTY = 0.5
"""Penalty for each gene copy to control the parsimony (0 for no penalty)."""


def estimate_cn(
    gene: Gene,
    coverage: Coverage,
    solver: str,
    gap: float = 0,
    fusion_penalty: float = LEFT_FUSION_PENALTY,
    user_solution=None,
    debug: Optional[str] = None,
) -> List[CNSolution]:
    """
    Estimate optimal copy number configurations given a gene and read data.

    :param gene: Gene instance.
    :param coverage: Read data coverage instance.
    :param solver:
        ILP solver. Check :obj:`aldy.lpinterface` for the list of supported solvers.
    :param gap:
        Relative optimality gap. Use non-zero values to allow non-optimal solutions.
        Default is 0 (reports only optimal solutions).
    :param fusion_penalty:
        Fusion penalty. Use higher values to avoid fusions.
        Default is 0.1.
    :param user_solution:
        User-specified list of copy number configurations.
        ILP solver will not run if this parameter is provided.
        Default is ``None``.
    :param debug:
        If set, create a "`debug`.cn.lp" file for debug purposes.
        Default is ``None``.

    :return: List of optimal copy number configurations.
    """

    log.debug("\n" + "*" * 80)

    if user_solution is not None:
        return [_parse_user_solution(gene, user_solution)]
    elif not gene.do_copy_number:
        basic = list(gene.cn_configs.keys())[0]
        return [_parse_user_solution(gene, [basic, basic])]
    else:
        # TODO: filter CN configs with non-present alleles
        max_observed_cn = 1 + max(
            int(round(coverage.region_coverage(gi, r)))
            for gi, g in enumerate(gene.regions)
            for r in g
        )
        region_cov = {
            r: (
                coverage.region_coverage(0, r),
                coverage.region_coverage(1, r) if len(gene.regions) > 1 else 0,
            )
            for r in gene.unique_regions
        }
        total_cov = sum(r0 + r1 for r0, r1 in region_cov.values())
        min_cov = min(
            sum(sum(v.values()) for v in gene.cn_configs[c].cn) for c in gene.cn_configs
        )
        if total_cov < min_cov / 2.0:
            raise AldyException(
                f"Coverage for {gene.name} too low for copy number calling."
            )

        configs = _filter_configs(gene, coverage)
        log.debug("[cn] candidates= {}", ", ".join(natsorted(configs)))
        log.debug("[cn] max_cn= {}", max_observed_cn)
        _print_coverage(gene, coverage)
        sol = solve_cn_model(
            gene,
            configs,
            max_observed_cn,
            region_cov,
            solver,
            gap,
            fusion_penalty,
            debug,
        )

        return sol


def solve_cn_model(
    gene: Gene,
    cn_configs: Dict[str, CNConfig],
    max_cn: int,
    region_coverage: Dict[str, Tuple[float, float]],
    solver: str,
    gap: float = 0,
    fusion_penalty: float = LEFT_FUSION_PENALTY,
    debug: Optional[str] = None,
) -> List[CNSolution]:
    """
    Solve the copy number estimation problem (an instance of closest vector problem).

    :param cn_configs: Available copy number configurations (vectors).
    :param max_cn: Maximum allowed copy number.
    :param region_coverage:
        Observed copy number of the main gene and the pseudogene for each genic region.
    :param solver:
        ILP solver to use.
        Check :obj:`aldy.lpinterface` for the list of supported solvers.
    :param gap:
        Relative optimality gap. Use non-zero values to allow non-optimal solutions.
        Default is 0 (reports only optimal solutions).
    :param fusion_penalty:
        Fusion penalty. Use higher values to avoid fusions.
        Default is 0.1.
    :param debug:
        If set, create a "`debug`.cn.lp" file for debug purposes.
        Default is ``None``.

    :return: List of optimal copy-number solutions.

    .. note::
        Please see `Aldy paper <https://www.nature.com/articles/s41467-018-03273-1>`_
        (section Methods/Copy number and structural variation estimation)
        for the detailed description of ILP model.
    """

    model = lpinterface.model("AldyCN", solver)
    debug_info = json["cn"]

    # List of CN configurations (a.k.a. structures):
    # dict of (`name`, `number`): structure, where multiple copies of the same `name`
    # get different numbers (binary LP transformation).
    # Note that `number` = {0, -1} represents full configuration with *all* pseudogenes.
    # Thus a diploid genome must have exactly *2* full configurations.
    # Any `number` > 0 describes only the main gene configuration
    # (a.k.a. "weak" configurations), and there can be many such configurations.
    structures = {(name, 0): structure for name, structure in cn_configs.items()}
    for a, ai in list(structures.keys()):
        structures[a, -1] = copy.deepcopy(structures[a, 0])
        if cn_configs[a].kind != CNConfigType.DEFAULT:
            continue
        for i in range(1, max_cn):
            structures[a, i] = copy.deepcopy(structures[a, 0])
            for g in range(1, len(structures[a, i].cn)):
                # If this is a pseudogene, remove a copy (create a "weak configuration")
                structures[a, i].cn[g] = {
                    r: v - 1 for r, v in structures[a, i].cn[g].items()
                }

    # Add a binary variable for each CN structure.
    VCN = {
        (a, ai): model.addVar(vtype="B", name=f"CN_{a}_{ai}") for a, ai in structures
    }

    # We assume diploid genome, so the number of haplotype-inducing configurations
    # must be 2.
    diplo_inducing = model.quicksum(VCN[a] for a in VCN if a[1] <= 0)
    model.addConstr(diplo_inducing <= 2, name="CDIPLO")
    model.addConstr(diplo_inducing >= 2, name="CDIPLO")

    # Ensure that we cannot link any allele to the whole-gene deletion.
    del_allele = gene.deletion_allele()
    if del_allele:
        for (a, ai), v in VCN.items():
            if a != del_allele:
                model.addConstr(v + VCN[del_allele, -1] <= 1, name=f"CDEL_{a}_{ai}")

    # Ensure that binary transformation is properly formed (i.e. A_i <= A_{i-1}).
    for a, ai in structures:
        # The second haplotype (-1) is only present if the first one (0) is there.
        if ai == -1:
            model.addConstr(VCN[a, ai] <= VCN[a, 0], name=f"CORD_{a}_{ai}")
        # Ignore 1, because A[1] = 1 && A[0] = 0 is valid (e.g. *13/*13+*1)
        elif ai > 1:
            model.addConstr(VCN[a, ai] <= VCN[a, ai - 1], name=f"CORD_{a}_{ai}")

    # Add error variables
    VERR = {}

    for r, (exp_cov0, exp_cov1) in region_coverage.items():
        debug_info["data"][r] = (exp_cov0, exp_cov1)
        expr = 0
        for s, structure in structures.items():
            if r in structure.cn[0]:
                expr += structure.cn[0][r] * VCN[s]
            if len(structure.cn) > 1 and r in structure.cn[1]:
                expr -= structure.cn[1][r] * VCN[s]
        VERR[r] = model.addVar(name=f"E_{r}", lb=-MAX_CN_ERROR, ub=MAX_CN_ERROR)
        model.addConstr(expr + VERR[r] <= exp_cov0 - exp_cov1, name=f"CCOV_{r}")
        model.addConstr(expr + VERR[r] >= exp_cov0 - exp_cov1, name=f"CCOV_{r}")

    # Objective: minimize the sum of absolute errors.
    # PCE_REGION (in CYP2D7) is penalized with an extra score as it is important
    # fusion marker.
    objective = model.abssum(VERR.values(), coeffs={"E_pce": PCE_PENALTY_COEFF})
    # Objective: also minimize the total number of present alleles (maximum parsimony)
    # Also penalize left fusions as they are not likely to occur.
    objective += model.quicksum(
        # MPICL requires only one term for each variable in the objective function,
        # thus this ugly expression:
        (fusion_penalty + PARSIMONY_PENALTY) * v
        if cn_configs[s].kind == CNConfigType.LEFT_FUSION
        else PARSIMONY_PENALTY * v
        for (s, k), v in VCN.items()
    )
    model.setObjective(objective)
    if debug:
        model.dump(f"{debug}.cn.lp")

    # Solve the model
    lookup = {model.varName(v): a for (a, ai), v in VCN.items()}
    result: dict = {}
    for status, opt, sol in model.solutions(gap):
        sol_tuple = sorted_tuple(lookup[v] for v in sol)
        # Because A[1] can be 1 while A[0] is 0, we can have biologically
        # homologous solutions
        if sol_tuple not in result:
            result[sol_tuple] = CNSolution(gene, opt, list(sol_tuple))
            log.debug(
                f"[cn] status= {status}; opt= {opt:.2f}; "
                + f"solution= {result[sol_tuple]}"
            )
    if not result:
        log.debug("[cn] solution= []")

    debug_info["sol"] = [dict(r.solution) for r in result.values()]
    return list(result.values())


def _filter_configs(gene: Gene, coverage: Coverage) -> Dict[str, CNConfig]:
    """
    Filter out low-quality mutations and copy number configurations
    that are not supported by the remaining mutations.
    """
    cov = coverage.filtered(
        lambda mut, cov, total, thres: Coverage.basic_filter(
            mut, cov, total, thres / MAX_CN
        )
    )
    configs = copy.deepcopy(gene.cn_configs)
    for an in natsorted(gene.cn_configs):
        if an not in gene.alleles:
            continue  # This is just a CN configuration w/o any mutations
        bad_alleles = []
        for a in gene.cn_configs[an].alleles:
            if any(cov[m] <= 0 for m in gene.alleles[a].func_muts):
                bad_alleles.append(a)
        if len(bad_alleles) == len(gene.cn_configs[an].alleles):
            log.trace(f"[cn] removing *{an} due to low support")
            del configs[an]
    return configs


def _print_coverage(gene: Gene, coverage: Coverage) -> None:
    """
    Pretty-print the region coverage.
    """
    log.debug("[cn] coverage=")
    gname = re.split(r"(\d.+)", gene.name)[1]
    pname = re.split(r"(\d.+)", gene.pseudogenes[0])[1] if gene.pseudogenes else ""
    log.debug(f"  {'':5} {gname:4} {pname:4}")
    for r in gene.regions[0]:
        g = coverage.region_coverage(0, r)
        if pname:
            p = coverage.region_coverage(1, r)
            if r in gene.unique_regions:
                log.debug(f"  {r:5} {g:4.1f} {p:4.1f} = {g - p:4.1f}")
            else:
                log.debug(f"  {r:5} {g:4.1f} {p:4.1f}")
        else:
            log.debug(f"  {r:5}: {g:4.1f}")


def _parse_user_solution(gene: Gene, sols: List[str]) -> CNSolution:
    """
    Parse user-provided copy number solutions.

    :param gene: Gene instance.
    :param sols: List of valid copy number configurations.

    :return: User-provided copy number solution.
    :raise: :obj:`aldy.common.AldyException` if a user-provided solution does not match
            the gene database.
    """
    for sol in sols:
        if sol not in gene.cn_configs:
            raise AldyException(
                "Given copy number solution contains unknown copy number configuration"
                + f"{sol}. Please run 'aldy show --gene {gene.name}' for the list the"
                + "valid configurations"
            )
    s = CNSolution(gene, 0, sols)  # type: ignore
    log.debug("[cn] result= {} (provided)", s)
    return s



# 786
# Aldy source: gene.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from typing import Tuple, Dict, List, Optional, Set, NamedTuple
from dataclasses import dataclass, field
from enum import Enum

import os
import yaml
import collections
from natsort import natsorted

from .common import (
    GRange,
    AldyException,
    allele_name,
    sorted_tuple,
    seq_to_amino,
    rev_comp,
    log,
    freezekey,
)


class Mutation(NamedTuple):
    """
    Mutation description. Immutable.

    :param pos: Reference genome position (0-based).
    :param op: Variant operation in HGVS-like format:
    - ``X>Y``:  a SNP from X to Y
    - ``insX``:  an insertion of X
    - ``delX``:  a deletion of X (X, Y are both of format ``[ACGT]+``).

    .. note:: Has custom printer (``__str__``).
    """

    pos: int
    op: str

    def __str__(self):
        return f"{self.pos + 1}.{self.op}"


@dataclass
class MinorAllele:
    """
    Minor allele description.

    :param name: Minor allale name.
    :param alt_name: List of alternative names.
    :param neutral_muts: Netural mutations that describe the minor allele.
    :param activity: Activity indicator (see PharmVar for details).
    :param evidence: Evidence indicator (see PharmVar for details).
    :param pharmvar: PharmVar ID.

    .. note:: Has custom printer (``__str__``).
    """

    name: str
    alt_name: Optional[str] = None
    neutral_muts: Set[Mutation] = field(default_factory=set)
    activity: Optional[str] = None
    evidence: Optional[str] = None
    pharmvar: Optional[str] = None

    def __str__(self):
        muts = ", ".join(map(str, self.neutral_muts))
        return f"Minor({self.name}; [{muts}])"


@dataclass
class MajorAllele:
    """
    Major allele description.

    :param name: Major allele name.
    :param cn_config: Copy-number configuration.
    :param func_muts: Functional mutations that describe the major allele.
    :param minors: Minor alleles (suballeles) that are derived from this major allele.
    """

    name: str
    cn_config: str = "1"
    func_muts: Set[Mutation] = field(default_factory=set)
    minors: Dict[str, MinorAllele] = field(default_factory=dict)

    def get_minor_mutations(self, minor: str):
        """Sequence of mutations that describe a minor allele."""
        for m in self.func_muts:
            yield m
        for m in self.minors[minor].neutral_muts:
            yield m


class CNConfigType(Enum):
    """Enumeration describing the type of a copy-number configuration"""

    DEFAULT = 0
    LEFT_FUSION = 1
    RIGHT_FUSION = 2
    DELETION = 3


@dataclass
class CNConfig:
    """
    Copy-number (CN) configuration description. Immutable.

    :param cn:
        Value of the expected region copy number in each gene.
        For example, ``cn[0]["e1"] == 1`` means that the exon 1 of the main gene (ID 0)
        has one copy (and thus should be present) in the configuration `cn`.
    :param kind: Type of the copy-number configuration.
    :param alleles: Allele IDs that have this CN configuration.
    :param description: Human-readable description (e.g. "deletion").

    .. note:: Has custom printer (``__str__``)
    """

    cn: List[Dict[str, int]]
    kind: CNConfigType
    alleles: Set[str]
    description: str = ""

    def __str__(self):
        cov = "|".join("".join(str(self.cn[g][r]) for r in self.cn[0]) for g in self.cn)
        alleles = " ".join(natsorted(self.alleles))
        return f"CNConfig({str(self.kind)[13:]}; vector={cov}; alleles=[{alleles}])"


@dataclass
class Gene:
    r"""
    Gene (and associated pseudogenes) description relative to a reference genome.

    :param name: Gene name (e.g. _CYP2D6_).
    :param version: Database version.
    :param ensembl: ENSEMBL gene ID.
    :param pharmvar: PharmVar ID.
    :param refseq: RefSeq sequence ID.
    :param seq: RefSeq sequence (typically \*1 allele).
    :param genome: Reference genome version (e.g. hg19).
    :param chr: Reference genome chromosome.
    :param strand: Reference genome strand of RefSeq sequence.
    :param chr_to_ref: Location mapping from the reference to the RefSeq sequence.
    :param ref_to_chr: Location mapping from the RefSeq sequence to the reference.
    :param pseudogenes: Pseudogene names. A pseudogene has ID greater than zero.
    :param regions:
        Collection of regions (names and ranges) for each gene in the database.
        Maps a gene ID to a dictionary that maps a gene region name
        (e.g. "e9" for exon 9) to its region in the reference genome.
        Gene 0 is the main gene.
    :param exons: List of RefSeq exon coordinates.
    :param aminoacid: Aminoacid sequence of the main gene.

    :param mutations:
        Maps ``(position, mutation_type)`` to a corresponding :obj:`Mutation`.
    :param mutation_info:
        Attributes of each mutation.
        These attributes currently consist:
        - function (None if the mutation is not functional),
        - rsID,
        - RefSeq position, and
        - RefSeq opcode.
    :param do_copy_number: Indicates if the gene is subject to structural variations.
    :param cn_configs:
        Copy-number configurations associated with the gene.
        `1` (akin to \*1) is the default CN config (no structural variations present).
    :param unique_regions: List of genic regions used for copy-number calling.
    :param alleles: Major star-alleles in the gene.
    :param common_tandems:
        List of common allele tandems. Used in diplotype assignment heuristics.
        For example, the fact that \*13 is always followed by \*1
        (encoded as ``('13', '1')``) will be used to group \*13 and \*1 together
        within the same haplotype (e.g. \*13+\*1).
    """

    # Basic information
    name: str
    version: str
    pharmvar: Optional[str]
    ensembl: Optional[str]

    # Structure
    refseq: str
    seq: str
    genome: str
    chr: str
    strand: int
    chr_to_ref: Dict[int, int]
    ref_to_chr: Dict[int, int]
    pseudogenes: List[str]
    regions: List[Dict[str, GRange]]
    exons: List[Tuple[int, int]]
    aminoacid: str

    # Mutations
    mutations: Dict[Tuple[int, str], Tuple[Optional[str], str, int, str]]

    # Copy number configurations
    do_copy_number: bool
    cn_configs: Dict[str, CNConfig]
    unique_regions: List[str]

    # Alleles
    alleles: Dict[str, MajorAllele]
    common_tandems: List[Tuple]

    # _lookup_range / _lookup_seq / _region_at

    def __init__(
        self,
        path: Optional[str],
        name: Optional[str] = None,
        yml: Optional[str] = None,
        genome: str = "hg19",
    ) -> None:
        """
        Initialize the Gene class with the database description
        specified in YML file ``path``.

        :param path: Location of YML file
        :param name: Gene name
        :param yml: YML file contents (used if ``path`` is ``None``).
        :param genome: Reference genome version
        :raise :obj:`aldy.common.AldyException`:: if a path is not set.
        """

        if path and os.path.exists(path):
            with open(path) as file:
                yml = file.read()
            if not name:
                name = os.path.split(path)[-1].split(".")[0].upper()

        if not yml or not name:
            raise AldyException("Either a path or a name should be given")

        yml = yaml.safe_load(yml)
        self.genome = genome
        self._init_basic(name, yml)
        self._init_regions(yml)
        self._init_alleles(yml)
        self._init_partials()

    def region_at(self, pos: int) -> Optional[Tuple[int, str]]:
        """
        :param pos: Position in the reference genome.
        :return: Gene ID and a region that covers the position.
        """
        return self._region_at.get(pos, None)

    def get_functional(self, mut, infer=True) -> Optional[str]:
        """
        :return: String describing the mutation effect if a mutation is functional;
                 otherwise None.
        """
        pos, op = mut
        if (pos, op) in self.mutations:
            return self.mutations[pos, op][0]

        def reverse_op(op: str) -> str:
            if ">" in op:
                l, r = op.split(">")
                return f"{rev_comp(l)}>{rev_comp(r)}"
            elif op[:3] == "ins":
                return f"ins{rev_comp(op[3:])}"
            elif op[:3] == "del":
                return f"del{rev_comp(op[3:])}"
            return op

        # Calculate based on aminoacid change
        pos = self.chr_to_ref[pos]
        if infer and any(s <= pos < e for s, e in self.exons):
            if ">" not in op:
                return "indel"
            if self.strand < 0:
                op = reverse_op(op)
            seq = "".join(
                self.seq[s:pos] + op[2] + self.seq[pos + 1 : e]
                if s <= pos < e
                else self.seq[s:e]
                for s, e in self.exons
            )
            amino = seq_to_amino(seq)
            if amino != self.aminoacid:
                pos = next(i for i, a in enumerate(amino) if a != self.aminoacid[i])
                return f"{self.aminoacid[pos]}{pos + 1}{amino[pos]}"
        return None

    def is_functional(self, mut, infer=True) -> bool:
        """
        :return: ``True`` if a mutation is functional.
        """
        return self.get_functional(mut, infer) is not None

    def get_rsid(self, *args, default=True) -> str:
        """
        :return: rsID if a mutation has it; otherwise "-".
            bool: ``True`` if a mutation is functional
            (i.e. does it affect the underlying aminoacid or not).
        """
        assert 1 <= len(args) <= 2
        if len(args) == 2:
            pos, op = args
        else:
            pos, op = args[0]
        res = "-"
        if (pos, op) in self.mutations:
            res = self.mutations[pos, op][1]
        return res if res != "-" or not default else f"{pos + 1}.{op}"

    def get_refseq(self, *args, from_atg=False) -> str:
        """
        Returns:
            bool: ``True`` if a mutation is functional
            (i.e. does it affect the underlying aminoacid or not).
        """
        assert 1 <= len(args) <= 2
        atg_start = self.exons[0][0]
        if len(args) == 2:
            pos, op = args
        else:
            pos, op = args[0]
        if (pos, op) in self.mutations:
            pos, op = self.mutations[pos, op][2:4]
            if from_atg:
                if pos >= atg_start:
                    pos = pos - atg_start + 1
                else:
                    pos = pos - atg_start
            return f"{pos}{op}"
        return "-"

    def deletion_allele(self) -> Optional[str]:
        """
        Returns:
            str, optional: The deletion allele ID. Can be ``None``
            if gene has no deletion allele.
        """
        try:
            return next(
                a
                for a, cn in self.cn_configs.items()
                if cn.kind == CNConfigType.DELETION
            )
        except StopIteration:
            return None

    def has_coverage(self, a: str, pos: int):
        """
        :return: ``True`` if a major allele `a` covers the mutation `m`.
        """
        m = self.region_at(pos)
        if m:
            return self.cn_configs[self.alleles[a].cn_config].cn[m[0]][m[1]] > 0
        return False

    def get_wide_region(self):
        mi = min(r.start for g in self.regions for r in g.values())
        ma = max(r.end for g in self.regions for r in g.values())
        return GRange(self.chr, mi, ma)

    def __contains__(self, i: int):
        return i in self.chr_to_ref

    def __getitem__(self, i):
        s, e = self._lookup_range
        if isinstance(i, slice):
            i, j = i.start, i.stop
            if j <= s or i >= e:
                return "N" * (j - i)
            loff = max(0, s - i)
            roff = max(0, j - e)
            if loff:
                i = s
            if roff:
                j = e
            return ("N" * loff) + self._lookup_seq[i - s : j - s] + ("N" * roff)
        else:
            assert isinstance(i, int)
            if not s <= i < e:
                return "N"
            return self._lookup_seq[i - s]

    def __str__(self):
        return f"Gene({self.name})"

    def __repr__(self):
        return f"Gene({self.name})"

    def _init_basic(self, gene: str, yml) -> None:
        """
        Read basic gene properties (``name``, ``seq`` and ``region``).

        All database YAML coordinates are indexed starting from 1.
        Aldy internally uses 0-based indexing.
        """
        self.name = yml["name"]
        self.version = f"{yml['version']} ({yml['generated']})"
        self.refseq = yml["reference"]["name"]
        self.pharmvar = yml.get("pharmvar", None)
        self.ensembl = yml.get("ensembl", None)

        self.seq = yml["reference"]["seq"].replace("\n", "")
        if "patches" in yml["reference"]:
            seq = list(self.seq)
            for [pos, nuc] in yml["reference"]["patches"]:
                seq[pos - 1] = nuc
            self.seq = "".join(self.seq)

        self.chr, start, end, strand, cigar = yml["reference"]["mappings"][self.genome]
        self.chr_to_ref = {}
        self.ref_to_chr = {}
        self.strand = 1 if strand == "+" else -1
        pos_ref = 0 if self.strand > 0 else (len(self.seq) - 1)
        pos_chr = start - 1
        for i in cigar.split():
            op, sz = i[0], int(i[1:])
            if op == "M":
                for idx in range(sz):
                    self.chr_to_ref[pos_chr + idx] = pos_ref + idx * self.strand
                    self.ref_to_chr[pos_ref + idx * self.strand] = pos_chr + idx
                pos_chr += sz
                pos_ref += sz * self.strand
            elif op == "I":
                pos_ref += sz * self.strand
            elif op == "D":
                pos_chr += sz
            else:
                raise AldyException("Invalid CIGAR string")

        self._lookup_range = (start - 1, end - 1)
        self._lookup_seq = "".join(
            (
                rev_comp(self.seq[self.chr_to_ref[i]])
                if self.strand < 0
                else self.seq[self.chr_to_ref[i]]
            )
            if i in self.chr_to_ref
            else "N"
            for i in range(*self._lookup_range)
        )

        self.exons = sorted((s - 1, e - 1) for [s, e] in yml["reference"]["exons"])
        self.aminoacid = seq_to_amino("".join(self.seq[s:e] for [s, e] in self.exons))

    def _init_regions(self, yml) -> None:
        """
        Calculate the genic regions and pseudogenes
        (``regions``, ``unique_regions`` and ``pseudogenes``).
        Prepare ``region_at()`` call.
        """

        self.regions = []
        # Each pseudogene is associated with an index > 0
        self.pseudogenes = []

        for i, g in enumerate(yml["structure"]["genes"]):
            if i > 0:
                self.pseudogenes.append(g)
            regions = {}
            num_exons = 0
            for name, coord in yml["structure"]["regions"][self.genome].items():
                if name[0] == "e" and name[1:].isdigit():  # this is exon
                    num_exons += 1
                if not coord[i * 2] <= coord[i * 2 + 1]:
                    raise AldyException(
                        f"Malformed YML file {self.name} (structure:regions:{name})"
                    )
                regions[name] = GRange(self.chr, coord[i * 2] - 1, coord[i * 2 + 1] - 1)
            for e in range(1, num_exons):  # fill introns
                r = [regions[f"e{e}"], regions[f"e{e + 1}"]][:: self.strand]
                regions[f"i{e}"] = GRange(self.chr, r[0][2], r[1][1])
            self.regions.append(
                dict(sorted(regions.items(), key=lambda x: x[1])[:: self.strand])
            )

        #: dict[int, (int, str)]:
        #: reverse lookup (gene, region) of gene regions given a location
        #: within the reference genome.
        for d in self.regions:
            regs = sorted((r.start, r.end) for r in d.values())
            for ri in range(1, len(regs)):
                if regs[ri - 1][1] < regs[ri][0]:
                    raise AldyException(
                        "Region {}-{} is not annotated", regs[ri - 1][1], regs[ri][0]
                    )
        self._region_at = {
            i: (g, r)
            for g, d in enumerate(self.regions)
            for r, rng in d.items()
            for i in range(rng.start, rng.end)
        }
        for i in self.chr_to_ref:
            if i not in self._region_at:
                raise AldyException(f"Position {i} not within a named region")
        self.unique_regions = yml["structure"]["cn_regions"]

    def _init_alleles(self, yml) -> None:
        """
        Initialize allele (``alleles``, ``common_tandems``)
        and copy number configurations (``cn_configs``).
        Initialize old notation lookup table (``old_notation``).

        Raises:
            :obj:`aldy.common.AldyException` if allele name cannot be found.
        """

        alleles = {}
        fusions_left = {}
        fusions_right = {}

        # allele ID of the deletion allele (i.e. whole gene is missing).
        deletion_allele = None

        self.mutations = {}
        for name, allele in yml["alleles"].items():
            name = allele_name(name)
            mutations: Set[Mutation] = set()
            if [self.name, "deletion"] in allele["mutations"]:
                deletion_allele = name
            else:
                for pos, op, *info in allele["mutations"]:
                    if pos in self.pseudogenes:
                        assert pos == self.pseudogenes[0]  # TODO: relax later
                        if op[-1] == "-":
                            fusions_left[name] = op[:-1]
                        else:
                            if op[-1] == "+":
                                op = op[:-1]
                            fusions_right[name] = op
                    else:
                        orig_op = op
                        rsid, function = info[0], info[1] if len(info) > 1 else None
                        if self.strand < 0:
                            if ">" in op:
                                l, r = op.split(">")
                                op = f"{rev_comp(l)}>{rev_comp(r)}"
                                pos = pos + len(l) - 1
                            elif op[:3] == "ins":
                                ins = op[3:]
                                while self.seq[pos - len(ins) : pos] == ins:
                                    pos -= len(ins)
                                pos += 1
                                op = f"ins{rev_comp(op[3:])}"
                            elif op[:3] == "del":
                                pos = pos + len(op) - 4
                                op = f"del{rev_comp(op[3:])}"
                        pos -= 1  # Cast to 0-based index
                        if pos not in self.ref_to_chr:
                            log.warn(
                                f"Ignoring {pos}.{op} in {name} (not in {self.refseq})"
                            )
                        elif self.region_at(self.ref_to_chr[pos]) is None:
                            log.warn(
                                f"Ignoring {pos}.{op} in {name} (not in named region)"
                            )
                        else:
                            mutations.add(Mutation(self.ref_to_chr[pos], op))
                            self.mutations.setdefault(
                                (self.ref_to_chr[pos], op),
                                (function, rsid, pos, orig_op),
                            )
            alt_name = allele.get("label", None)
            if alt_name:
                alt_name = allele_name(alt_name)
            alleles[name] = MinorAllele(
                name,
                alt_name,
                mutations,
                allele.get("activity", None),
                allele.get("evidence", None),
                allele.get("pharmvar", None),
            )

        # TODO:
        self.common_tandems: List[tuple] = []
        if "tandems" in yml["structure"]:
            self.common_tandems = [tuple(y) for y in yml["structure"]["tandems"]]

        # Set copy number configurations
        # TODO: currently fusions only cover the space between the main gene and
        # the first pseudogene. Multi-pseudogene fusions are not supported.

        self.do_copy_number = bool(
            deletion_allele or len(fusions_left) or len(fusions_right)
        )
        self.cn_configs: Dict[str, CNConfig] = dict()

        for i, _ in enumerate(self.pseudogenes):
            if list(self.regions[i + 1].keys()) != list(self.regions[0].keys()):
                raise AldyException("Invalid database structure")
        rank = {a: ai for ai, a in enumerate(self.regions[0])}
        inverse_cn: Dict[tuple, str] = dict()
        # Left fusions are PSEUDOGENE + GENE fusions
        for a, brk in fusions_left.items():
            cn = [
                {r: int(rank[r] >= rank[brk]) for r in self.regions[0]},
                {r: int(rank[r] < rank[brk]) for r in self.regions[1]},
            ]
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn,
                    CNConfigType.LEFT_FUSION,
                    {a},
                    f"{self.pseudogenes[0]} fusion until {brk}",
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)
        # Deletion is a special kind of left fusion
        if deletion_allele is not None:
            cn = [{r: 0 for r in self.regions[0]}]
            if len(self.pseudogenes) > 0:
                cn.append({r: 1 for r in self.regions[1]})
            self.cn_configs[deletion_allele] = CNConfig(
                cn,
                CNConfigType.DELETION,
                {deletion_allele},
                f"{self.name} deletion",
            )
        # Right fusions GENE + PSEUDOGENE + whole copy of PSEUDOGENE fusions
        for a, brk in fusions_right.items():
            cn = [
                {r: int(rank[r] < rank[brk]) for r in self.regions[0]},
                {r: 1 + int(rank[r] >= rank[brk]) for r in self.regions[1]},
            ]
            key = freezekey(cn)
            if key not in inverse_cn:
                self.cn_configs[a] = CNConfig(
                    cn,
                    CNConfigType.RIGHT_FUSION,
                    {a},
                    f"{self.pseudogenes[0]} conservation after {brk}",
                )
                inverse_cn[key] = a
            else:
                self.cn_configs[inverse_cn[key]].alleles.add(a)

        # Normal CN case
        used_alleles = {a for cn in self.cn_configs.values() for a in cn.alleles}
        has_pseudogenes = len(self.pseudogenes) > 0
        default_cn = [
            {r: 1 for r in self.regions[g]} for g in range(1 + has_pseudogenes)
        ]
        self.cn_configs = {min(v.alleles): v for k, v in self.cn_configs.items()}
        self.cn_configs["1"] = CNConfig(
            default_cn,
            CNConfigType.DEFAULT,
            set(alleles.keys()) - used_alleles,
            "Standard copy-number configuration",
        )

        # Set CN of "empty" regions to zero (e.g. CYP2D6.pce)
        for _, conf in self.cn_configs.items():
            for g, r in enumerate(conf.cn):
                for rg in r:
                    if self.regions[g][rg].end - self.regions[g][rg].start <= 0:
                        r[rg] = 0

        # Set up major and minor allele structures
        alleles_inverse: Dict[tuple, set] = collections.defaultdict(set)
        for a in alleles:
            cn_config = next(
                cn for cn, conf in self.cn_configs.items() if a in conf.alleles
            )
            fn_muts = sorted(
                m for m in alleles[a].neutral_muts if self.is_functional(m)
            )
            alleles_inverse[cn_config, tuple(fn_muts)].add(a)

        # Ensure that each distinct major allele has unique prefix
        used_names: Dict[str, int] = {}
        new_names = {}
        changed_cn_configs = {}
        for key, minors in alleles_inverse.items():
            an = min(minors)
            name = an.split(".")[0]  # remove ".0XX" suffix
            if name in used_names:  # special case
                name = alleles[an].alt_name if alleles[an].alt_name else an
            if name in used_names:  # Append letter
                used_names[name] += 1
                log.info(f"Renaming {name} -> {name}:{used_names[name]}")
                name += f":{used_names[name]}"
                used_names[name] = 1
            else:
                used_names[name] = 1
            if an in self.cn_configs and an != name:
                self.cn_configs[name] = self.cn_configs[an]
                changed_cn_configs[an] = name
                del self.cn_configs[an]
            new_names[key] = name
        self.alleles: Dict[str, MajorAllele] = dict()
        for key, minors in alleles_inverse.items():
            self.alleles[new_names[key]] = MajorAllele(
                name=new_names[key],
                cn_config=changed_cn_configs.get(key[0], key[0]),
                func_muts=set(key[1]),
                minors={
                    sa: MinorAllele(
                        sa,
                        alleles[sa].alt_name,
                        set(alleles[sa].neutral_muts) - set(key[1]),
                        alleles[sa].activity,
                        alleles[sa].evidence,
                        alleles[sa].pharmvar,
                    )
                    for sa in natsorted(minors)
                },
            )

    def _init_partials(self) -> None:
        """
        Construct "partial" major alleles.
        If a major allele is cut in half by a fusion, we will create a "new" major
        allele that contains the functional mutations that have survived
        the fusion event.

        Note:
            This is currently supported only for left fusions.
        """

        def preserved_mutations(f, m):
            def filter_f(m):
                m = self.region_at(m.pos)
                if m:
                    return self.cn_configs[f].cn[m[0]][m[1]] > 0
                return False

            return set(filter(filter_f, m))

        for f in filter(
            lambda x: self.cn_configs[x].kind == CNConfigType.LEFT_FUSION,
            self.cn_configs,
        ):
            # Do not extend left fusions that are already defined
            # by some functional mutations
            # HACK: This is hacky but works well for now
            if len(self.alleles[f].func_muts) > 0:
                continue
            add: Dict[tuple, MajorAllele] = {}
            for an, a in self.alleles.items():
                # We only make partial major alleles from non-fused major alleles
                if a.cn_config != "1":
                    continue
                new_name = f"{f}#{an}"
                new_muts = preserved_mutations(f, a.func_muts)
                key = sorted_tuple(new_muts)
                if key in add:
                    add[key].minors.update(
                        {
                            f"{f}#{san}": MinorAllele(
                                f"{f}#{san}",
                                None,
                                preserved_mutations(f, sa.neutral_muts),
                                sa.activity,
                                sa.evidence,
                                sa.pharmvar,
                            )
                            for san, sa in a.minors.items()
                        }
                    )
                else:
                    add[key] = MajorAllele(
                        name=new_name,
                        cn_config=f,
                        func_muts=new_muts,
                        minors={
                            f"{f}#{san}": MinorAllele(
                                f"{f}#{san}",
                                neutral_muts=preserved_mutations(f, sa.neutral_muts),
                            )
                            for san, sa in a.minors.items()
                        },
                    )

            # Remove fusion (will be replaced at least by allele "1#{f}")
            del self.alleles[f]
            self.alleles.update({a.name: a for a in add.values()})

        for an, a in self.alleles.items():
            # Clean up minor alleles (as many might be identical after a fusion).
            # Put a reference to the cleaned-up alleles in ``alt_name`` field.
            minors: Dict[tuple, List[str]] = collections.defaultdict(list)
            for s in a.minors:
                key = sorted_tuple(a.minors[s].neutral_muts)
                minors[key].append(s)
            self.alleles[an] = MajorAllele(
                self.alleles[an].name,
                self.alleles[an].cn_config,
                self.alleles[an].func_muts,
                {
                    min(sa): MinorAllele(
                        min(sa),
                        a.minors[min(sa)].alt_name,
                        set(nm),
                        a.minors[min(sa)].activity,
                        a.minors[min(sa)].evidence,
                        a.minors[min(sa)].pharmvar,
                    )
                    for nm, sa in minors.items()
                },
            )

        # TODO: prune identical post-fusion alleles
        # (e.g. *2 after *13A and *13B is the same--- no need to have 2 items).

        for _, v in self.cn_configs.items():
            v.alleles.clear()
        for a in self.alleles.values():
            self.cn_configs[a.cn_config].alleles.add(a.name)
